# Copyright 2017, 2019 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import enum
import time
import abc
import asyncio
import warnings
from typing import (Generic, TypeVar, Type, List, Tuple, Mapping, Iterable,
                    KeysView, ValuesView, ItemsView,
                    Iterator, Optional, Any, Union, Callable, cast, overload)
# Imports only used for type comments, otherwise unused
from typing import Set, Dict     # noqa: F401

from . import core


_T = TypeVar('_T')


class Reading(Generic[_T]):
    """Sensor reading

    Parameters
    ----------
    timestamp
        The UNIX timestamp at which the sensor value was determined
    status
        Sensor status at `timestamp`
    value
        Sensor value at `timestamp`
    """
    # Note: can't use slots in 3.5 due to https://bugs.python.org/issue28790

    def __init__(self, timestamp: float, status: 'Sensor.Status', value: _T) -> None:
        self.timestamp = timestamp
        self.status = status
        self.value = value


def _default_status_func(value) -> 'Sensor.Status':
    return Sensor.Status.NOMINAL


class Sensor(Generic[_T]):
    """A sensor in a :class:`DeviceServer`.

    A sensor has some static configuration (name, description, units etc) and
    dynamic state consisting of a :class:`Reading` (value, status and
    timestamp). Other code can attach observers to the sensor to be informed of
    updates.

    Parameters
    ----------
    sensor_type
        The type of the sensor.
    name
        Sensor name
    description
        More detailed explanation of the sensor
    units
        Physical units of the sensor
    default
        Initial value of the sensor. When setting this, it may be desirable to
        specify `initial_status` too.
    initial_status
        Initial status of the sensor
    status_func
        Function that maps a value to a status in :meth:`set_value` if none is given.
        The default is a function that always returns NOMINAL.
    auto_strategy
        Sampling strategy to use when a client requests the ``auto`` strategy.
        The default is to send all updates of the value to the client
        immediately.
    auto_strategy_parameters
        Parameters to use with `auto_strategy`. They must be already-decoded values.
    """

    class Status(enum.Enum):
        UNKNOWN = 0
        NOMINAL = 1
        WARN = 2
        ERROR = 3
        FAILURE = 4
        UNREACHABLE = 5
        INACTIVE = 6

        def valid_value(self) -> bool:
            """True if this state is one where the value provided is valid."""
            return self in {Sensor.Status.NOMINAL, Sensor.Status.WARN, Sensor.Status.ERROR}

    def __init__(self, sensor_type: Type[_T],
                 name: str,
                 description: str = None,
                 units: str = '',
                 default: _T = None,
                 initial_status: Status = Status.UNKNOWN,
                 *,
                 status_func: Callable[[_T], Status] = _default_status_func,
                 auto_strategy: Optional['SensorSampler.Strategy'] = None,
                 auto_strategy_parameters: Iterable[Any] = ()) -> None:
        self.stype = sensor_type
        type_info = core.get_type(sensor_type)
        self.type_name = type_info.name
        self._observers = set()           # type: Set[Callable[[Sensor[_T], Reading[_T]], None]]
        self.name = name
        self.description = description
        self.units = units
        self.status_func = status_func
        if default is None:
            value = type_info.default(sensor_type)   # type: _T
        else:
            value = default
        self._reading = Reading(time.time(), initial_status, value)
        if auto_strategy is None:
            self.auto_strategy = SensorSampler.Strategy.AUTO
        else:
            self.auto_strategy = auto_strategy
        self.auto_strategy_parameters = tuple(auto_strategy_parameters)
        # TODO: should validate the parameters against the strategy.

    def notify(self, reading: Reading[_T]) -> None:
        """Notify all observers of changes to this sensor.

        Users should not usually call this directly. It is called automatically
        by :meth:`set_value`.
        """
        for observer in self._observers:
            observer(self, reading)

    def set_value(self, value: _T, status: Status = None,
                  timestamp: float = None) -> None:
        """Set the current value of the sensor.

        Parameters
        ----------
        value
            The value of the sensor (the type should be appropriate to the
            sensor's type).
        status
            Whether the value represents an error condition or not. If not
            given, the `status_func` given to the constructor is used to
            determine the status from the value.
        timestamp
            The time at which the sensor value was determined (seconds).
            If not given, it defaults to :func:`time.time`.
        """
        if timestamp is None:
            timestamp = time.time()
        if status is None:
            status = self.status_func(value)
        reading = Reading(timestamp, status, value)
        self._reading = reading
        self.notify(reading)

    @property
    def value(self) -> _T:
        """The current value of the sensor.

        Modifying it invokes :meth:`set_value`.
        """
        return self.reading.value

    @value.setter
    def value(self, value: _T) -> None:
        self.set_value(value)

    @property
    def timestamp(self) -> float:
        return self.reading.timestamp

    @property
    def status(self) -> Status:
        return self.reading.status

    @property
    def reading(self) -> Reading[_T]:
        return self._reading

    @property
    def params(self) -> List[bytes]:
        if self.type_name == 'discrete':
            return [core.encode(value) for value in cast(Iterable, self.stype)]
        else:
            return []

    def attach(self, observer: Callable[['Sensor[_T]', Reading[_T]], None]) -> None:
        self._observers.add(observer)

    def detach(self, observer: Callable[['Sensor[_T]', Reading[_T]], None]) -> None:
        self._observers.discard(observer)


class SensorSampler(Generic[_T], metaclass=abc.ABCMeta):
    class Strategy(enum.Enum):
        NONE = 0
        AUTO = 1
        PERIOD = 2
        EVENT = 3
        DIFFERENTIAL = 4
        EVENT_RATE = 5
        DIFFERENTIAL_RATE = 6

    def __init__(self, sensor: Sensor[_T], observer: Callable[[Sensor[_T], Reading[_T]], None],
                 loop: asyncio.AbstractEventLoop,
                 difference: Optional[_T] = None,
                 shortest: core.Timestamp = core.Timestamp(0),
                 longest: core.Timestamp = None,
                 *, always_update: bool = False, is_auto: bool = False) -> None:
        if longest is not None:
            self.longest = float(longest)  # type: Optional[float]
            if self.longest <= 0:
                raise ValueError('period must be positive')
        else:
            self.longest = None
        self.shortest = float(shortest)
        self.sensor = sensor            # type: Optional[Sensor[_T]]
        self.observer = observer        # type: Optional[Callable[[Sensor[_T], Reading[_T]], None]]
        self.difference = difference
        self.always_update = always_update
        self.is_auto = is_auto
        self.loop = loop
        self._callback_handle = None    # type: Optional[asyncio.Handle]
        self._last_time = 0.0
        self._last_value = None         # type: Optional[_T]
        self._last_status = None        # type: Optional[Sensor.Status]
        self._changed = False
        self.sensor.attach(self._receive_update)
        self._send_update(loop.time(), sensor.reading)

    def __del__(self) -> None:
        if getattr(self, 'sensor', None) is not None:
            warnings.warn('unclosed SensorSampler {!r}'.format(self), ResourceWarning)
            if not self.loop.is_closed():
                self.loop.call_soon_threadsafe(self.close)

    def _clear_callback(self) -> None:
        if self._callback_handle is not None:
            self._callback_handle.cancel()
            self._callback_handle = None

    def _send_update(self, sched_time: float, reading: Optional[Reading[_T]]) -> None:
        assert self.sensor is not None
        assert self.observer is not None
        if reading is None:
            reading = self.sensor.reading
        self.observer(self.sensor, reading)
        self._last_time = sched_time
        self._last_value = reading.value
        self._last_status = reading.status
        self._changed = False
        self._clear_callback()
        if self.longest is not None:
            next_time = max(self.loop.time(), sched_time + self.longest)
            self._callback_handle = self.loop.call_at(
                next_time, self._send_update, next_time, None)

    def _receive_update(self, sensor: Sensor[_T], reading: Reading[_T]) -> None:
        if self._changed:
            # We already know the value changed, we're waiting for time-based callback
            return

        if self.always_update:
            changed = True
        elif self.difference is None:
            changed = reading.value != self._last_value
        else:
            assert sensor.stype in (int, float)
            changed = abs(cast(Any, reading.value) - self._last_value) > self.difference
        if reading.status != self._last_status:
            changed = True

        if changed:
            self._changed = True
            self._clear_callback()
            sched_time = self._last_time + self.shortest
            now = self.loop.time()
            if not self.shortest or now >= sched_time:
                self._send_update(now, reading)
            else:
                self._callback_handle = self.loop.call_at(
                    sched_time, self._send_update, sched_time, None)

    def close(self) -> None:
        """Stop monitoring the sensor.

        This should be called when the sampler is no longer needed. It is not
        valid to call any methods on the sampler after this.
        """
        self._clear_callback()
        if self.sensor is not None:
            self.sensor.detach(self._receive_update)
            self.sensor = None
        self.observer = None

    @abc.abstractmethod
    def _parameters(self) -> tuple:
        pass       # pragma: no cover

    def parameters(self) -> tuple:
        """Return the parameters with which the sensor was created."""
        if self.is_auto:
            return (SensorSampler.Strategy.AUTO,)
        else:
            return self._parameters()

    @classmethod
    def factory(cls, sensor: Sensor[_T], observer: Callable[[Sensor[_T], Reading[_T]], None],
                loop: asyncio.AbstractEventLoop,
                strategy: 'SensorSampler.Strategy', *args: bytes) -> Optional['SensorSampler[_T]']:
        classes_types = {
            cls.Strategy.NONE: (None, []),
            cls.Strategy.AUTO: (_SensorSamplerEventAlways, []),
            cls.Strategy.PERIOD: (_SensorSamplerPeriod, [core.Timestamp]),
            cls.Strategy.EVENT: (_SensorSamplerEvent, []),
            cls.Strategy.DIFFERENTIAL: (_SensorSamplerDifferential, [sensor.stype]),
            cls.Strategy.EVENT_RATE: (_SensorSamplerEventRate, [core.Timestamp, core.Timestamp]),
            cls.Strategy.DIFFERENTIAL_RATE:
                (_SensorSamplerDifferentialRate, [sensor.stype, core.Timestamp, core.Timestamp])
        }   # type: Dict[SensorSampler.Strategy, Tuple[Optional[Type[SensorSampler]], List[Type]]]

        if strategy == cls.Strategy.AUTO:
            strategy = sensor.auto_strategy
            decoded_args = sensor.auto_strategy_parameters
            out_cls = classes_types[strategy][0]
            is_auto = True
        else:
            if strategy in (cls.Strategy.DIFFERENTIAL, cls.Strategy.DIFFERENTIAL_RATE):
                if sensor.stype not in (int, float):
                    raise TypeError(
                        'differential strategies only valid for integer and float sensors')
            out_cls, types = classes_types[strategy]
            if len(types) != len(args):
                raise ValueError('expected {} strategy arguments, found {}'.format(
                    len(types), len(args)))
            decoded_args = tuple(core.decode(type_, arg) for type_, arg in zip(types, args))
            is_auto = False

        if out_cls is None:
            return None
        else:
            return out_cls(sensor, observer, loop, *decoded_args, is_auto=is_auto)


class _SensorSamplerEventAlways(SensorSampler[_T]):
    def __init__(self, sensor: Sensor[_T], observer: Callable[[Sensor[_T], Reading[_T]], None],
                 loop: asyncio.AbstractEventLoop, *, is_auto: bool) -> None:
        super().__init__(sensor, observer, loop, always_update=True, is_auto=is_auto)

    def _parameters(self) -> Tuple[SensorSampler.Strategy]:
        return (SensorSampler.Strategy.AUTO,)


class _SensorSamplerPeriod(SensorSampler[_T]):
    def __init__(self, sensor: Sensor[_T], observer: Callable[[Sensor[_T], Reading[_T]], None],
                 loop: asyncio.AbstractEventLoop, period: core.Timestamp,
                 *, is_auto: bool) -> None:
        super().__init__(sensor, observer, loop, shortest=period, longest=period, is_auto=is_auto)

    def _parameters(self) -> Tuple[SensorSampler.Strategy, core.Timestamp]:
        return (SensorSampler.Strategy.PERIOD, core.Timestamp(self.shortest))


class _SensorSamplerEvent(SensorSampler[_T]):
    def _parameters(self) -> Tuple[SensorSampler.Strategy]:
        return (SensorSampler.Strategy.EVENT,)


class _SensorSamplerDifferential(SensorSampler[_T]):
    def __init__(self, sensor: Sensor[_T], observer: Callable[[Sensor[_T], Reading[_T]], None],
                 loop: asyncio.AbstractEventLoop, difference: _T,
                 *, is_auto: bool) -> None:
        super().__init__(sensor, observer, loop, difference=difference, is_auto=is_auto)

    def _parameters(self) -> Tuple[SensorSampler.Strategy, _T]:
        assert self.difference is not None      # To keep mypy happy
        return (SensorSampler.Strategy.DIFFERENTIAL, self.difference)


class _SensorSamplerEventRate(SensorSampler[_T]):
    def __init__(self, sensor: Sensor[_T], observer: Callable[[Sensor[_T], Reading[_T]], None],
                 loop: asyncio.AbstractEventLoop,
                 shortest: core.Timestamp,
                 longest: core.Timestamp,
                 *, is_auto: bool) -> None:
        super().__init__(sensor, observer, loop, shortest=shortest, longest=longest,
                         is_auto=is_auto)

    def _parameters(self) -> Tuple[SensorSampler.Strategy, core.Timestamp, core.Timestamp]:
        # assertions to keep mypy happy
        assert self.shortest is not None
        assert self.longest is not None
        return (SensorSampler.Strategy.EVENT_RATE,
                core.Timestamp(self.shortest),
                core.Timestamp(self.longest))


class _SensorSamplerDifferentialRate(SensorSampler[_T]):
    def __init__(self, sensor: Sensor[_T], observer: Callable[[Sensor[_T], Reading[_T]], None],
                 loop: asyncio.AbstractEventLoop,
                 difference: _T,
                 shortest: core.Timestamp,
                 longest: core.Timestamp,
                 *, is_auto: bool) -> None:
        super().__init__(sensor, observer, loop,
                         difference=difference, shortest=shortest, longest=longest, is_auto=is_auto)

    def _parameters(self) -> Tuple[SensorSampler.Strategy, _T, core.Timestamp, core.Timestamp]:
        # assertions to keep mypy happy
        assert self.difference is not None
        assert self.shortest is not None
        assert self.longest is not None
        return (SensorSampler.Strategy.DIFFERENTIAL_RATE,
                self.difference,
                core.Timestamp(self.shortest),
                core.Timestamp(self.longest))


class SensorSet(Mapping[str, Sensor]):
    """A dict-like and set-like collection of sensors.

    It is possible to monitor for removal of sensors using
    :meth:`add_remove_callback`.
    """

    class _Sentinel(enum.Enum):
        """Internal enum used to signal that no default is provided to pop"""
        NO_DEFAULT = 0

    def __init__(self) -> None:
        self._sensors = {}             # type: Dict[str, Sensor]
        self._add_callbacks = []       # type: List[Callable[[Sensor], None]]
        self._remove_callbacks = []    # type: List[Callable[[Sensor], None]]

    def _removed(self, s: Sensor) -> None:
        """Call all the remove callbacks."""
        for callback in list(self._remove_callbacks):
            callback(s)

    def _added(self, s: Sensor) -> None:
        """Call all the add callbacks."""
        for callback in list(self._add_callbacks):
            callback(s)

    def add_remove_callback(self, callback: Callable[[Sensor], None]) -> None:
        """Add a callback that will be passed any sensor removed from the set."""
        self._remove_callbacks.append(callback)

    def add_add_callback(self, callback: Callable[[Sensor], None]) -> None:
        """Add a callback that will be passed any sensor added to the set."""
        self._add_callbacks.append(callback)

    def remove_remove_callback(self, callback: Callable[[Sensor], None]) -> None:
        """Remove a callback registered with :meth:`add_remove_callback`."""
        self._remove_callbacks.remove(callback)

    def remove_add_callback(self, callback: Callable[[Sensor], None]) -> None:
        """Remove a callback registered with :meth:`add_add_callback`."""
        self._add_callbacks.remove(callback)

    def add(self, elem: Sensor) -> None:
        if elem.name in self._sensors:
            if self._sensors[elem.name] is not elem:
                del self[elem.name]
            else:
                return
        self._sensors[elem.name] = elem
        self._added(elem)

    def remove(self, elem: Sensor) -> None:
        if elem not in self:
            raise KeyError(elem.name)
        del self[elem.name]

    def discard(self, elem: Sensor) -> None:
        try:
            self.remove(elem)
        except KeyError:
            pass

    def clear(self) -> None:
        while self._sensors:
            self.popitem()

    def popitem(self) -> Tuple[str, Sensor]:
        name, value = self._sensors.popitem()
        self._removed(value)
        return name, value

    def pop(self, key: str,
            default: Union[Sensor, None, _Sentinel] = _Sentinel.NO_DEFAULT) -> Optional[Sensor]:
        if key not in self._sensors:
            if isinstance(default, self._Sentinel):
                raise KeyError(key)
            else:
                return default
        else:
            s = self._sensors.pop(key)
            self._removed(s)
            return s

    def __delitem__(self, key: str) -> None:
        s = self._sensors.pop(key)
        self._removed(s)

    def __getitem__(self, name: str) -> Sensor:
        return self._sensors[name]

    @overload
    def get(self, name: str) -> Optional[Sensor]: ...

    @overload     # noqa: F811
    def get(self, name: str, default: Union[Sensor, _T]) -> Union[Sensor, _T]: ...

    def get(self, name: str, default: object = None) -> object:    # noqa: F811
        return self._sensors.get(name, default)

    def __contains__(self, s: object) -> bool:
        if isinstance(s, Sensor):
            return s.name in self._sensors and self._sensors[s.name] is s
        else:
            return s in self._sensors

    def __len__(self) -> int:
        return len(self._sensors)

    def __bool__(self) -> bool:
        return bool(self._sensors)

    def __iter__(self) -> Iterator[str]:
        return iter(self._sensors)

    def keys(self) -> KeysView[str]:
        return self._sensors.keys()

    def values(self) -> ValuesView[Sensor]:
        return self._sensors.values()

    def items(self) -> ItemsView[str, Sensor]:
        return self._sensors.items()

    def copy(self) -> Dict[str, Sensor]:
        return self._sensors.copy()

    __hash__ = None     # type: ignore     # mypy can't handle this

    add.__doc__ = set.add.__doc__
    remove.__doc__ = set.remove.__doc__
    discard.__doc__ = set.discard.__doc__
    clear.__doc__ = dict.clear.__doc__
    popitem.__doc__ = dict.popitem.__doc__
    pop.__doc__ = dict.pop.__doc__
    get.__doc__ = dict.get.__doc__
    keys.__doc__ = dict.keys.__doc__
    values.__doc__ = dict.values.__doc__
    items.__doc__ = dict.items.__doc__
    copy.__doc__ = dict.copy.__doc__



# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import asyncio
import logging
from typing import Optional     # noqa: F401

import asynctest

from aiokatcp.core import Message, KatcpSyntaxError
from aiokatcp.connection import read_message, Connection
from .test_utils import timelimit


class TestReadMessage(asynctest.TestCase):
    forbid_get_event_loop = True

    async def test_read(self) -> None:
        data = b'?help[123] foo\n#log info msg\n \t\n\n!help[123] bar\n\n'
        reader = asyncio.StreamReader(loop=self.loop)
        reader.feed_data(data)
        reader.feed_eof()
        msg = await read_message(reader)
        self.assertEqual(msg, Message.request('help', 'foo', mid=123))
        msg = await read_message(reader)
        self.assertEqual(msg, Message.inform('log', 'info', 'msg'))
        msg = await read_message(reader)
        self.assertEqual(msg, Message.reply('help', 'bar', mid=123))
        msg = await read_message(reader)
        self.assertIsNone(msg)

    async def test_read_overrun(self) -> None:
        data = b'!foo a_string_that_doesnt_fit_in_the_buffer\n!foo short_string\n'
        reader = asyncio.StreamReader(limit=25, loop=self.loop)
        reader.feed_data(data)
        reader.feed_eof()
        with self.assertRaises(KatcpSyntaxError):
            await read_message(reader)
        msg = await read_message(reader)
        self.assertEqual(msg, Message.reply('foo', 'short_string'))

    async def test_read_overrun_eof(self) -> None:
        data = b'!foo a_string_that_doesnt_fit_in_the_buffer'
        reader = asyncio.StreamReader(limit=25, loop=self.loop)
        reader.feed_data(data)
        reader.feed_eof()
        with self.assertRaises(KatcpSyntaxError):
            await read_message(reader)
        msg = await read_message(reader)
        self.assertIsNone(msg)

    async def test_read_partial(self) -> None:
        data = b'!foo nonewline'
        reader = asyncio.StreamReader(loop=self.loop)
        reader.feed_data(data)
        reader.feed_eof()
        with self.assertRaises(KatcpSyntaxError):
            await read_message(reader)


@timelimit
class TestConnection(asynctest.TestCase):
    async def setUp(self) -> None:
        self.writer = None       # type: Optional[asyncio.StreamWriter]
        self.reader = None       # type: Optional[asyncio.StreamReader]
        #: If non-None, _ok_reply waits on it before sending the reply
        self.ok_wait = None      # type: Optional[asyncio.Semaphore]
        #: Released by _ok_reply once the reply is sent
        self.ok_done = asyncio.Semaphore(0, loop=self.loop)
        #: Set ready once the connection has been established
        self._ready = asyncio.Event(loop=self.loop)
        self.server = await asyncio.start_server(
            self._client_connected_cb, '127.0.0.1', 0, loop=self.loop)
        host, port = self.server.sockets[0].getsockname()    # type: ignore
        self.owner = asynctest.MagicMock()
        self.owner.loop = self.loop
        self.owner.handle_message = asynctest.CoroutineMock(side_effect=self._ok_handler)
        self.remote_reader, self.remote_writer = await asyncio.open_connection(
            host, port, loop=self.loop)
        # Ensure the server side of the connection is ready
        await self._ready.wait()

    async def tearDown(self) -> None:
        self.server.close()
        await self.server.wait_closed()
        self.remote_writer.close()
        if self.writer:
            self.writer.close()

    def make_connection(self, is_server: bool) -> Connection:
        assert self.reader    # keeps mypy happy
        assert self.writer
        conn = Connection(self.owner, self.reader, self.writer, is_server)
        self.addCleanup(conn.wait_closed)
        self.addCleanup(conn.close)
        return conn

    def _client_connected_cb(self, reader: asyncio.StreamReader,
                             writer: asyncio.StreamWriter) -> None:
        self.reader = reader
        self.writer = writer
        self._ready.set()

    async def _ok_reply(self, conn, msg):
        if self.ok_wait:
            await self.ok_wait.acquire()
        conn.write_message(Message.reply_to_request(msg, 'ok'))
        await conn.drain()
        self.ok_done.release()

    async def _ok_handler(self, conn, msg):
        self.loop.create_task(self._ok_reply(conn, msg))

    async def test_write_message(self) -> None:
        conn = self.make_connection(True)
        conn.write_message(Message.reply('ok', mid=1))
        await conn.drain()
        line = await self.remote_reader.readline()
        self.assertEqual(line, b'!ok[1]\n')

    async def test_run(self) -> None:
        conn = self.make_connection(True)
        self.remote_writer.write(b'?watchdog[2]\n')
        await self.ok_done.acquire()
        self.owner.handle_message.assert_called_with(conn, Message.request('watchdog', mid=2))
        reply = await self.remote_reader.readline()
        self.assertEqual(reply, b'!watchdog[2] ok\n')
        # Check that it exits when the client disconnects its write end
        self.remote_writer.write_eof()
        # We need to give the packets time to go through the system
        await conn.wait_closed()

    async def test_disconnected(self) -> None:
        conn = self.make_connection(True)
        # Don't send the reply until the socket is closed
        self.ok_wait = asyncio.Semaphore(0, loop=self.loop)
        self.remote_writer.write(b'?watchdog[2]\n?watchdog[3]\n?watchdog[4]\n')
        self.remote_writer.close()
        await asynctest.exhaust_callbacks(self.loop)
        self.ok_wait.release()
        self.ok_wait.release()
        with self.assertLogs('aiokatcp.connection', logging.WARNING) as cm:
            # Wait for first two watchdogs and the close to go through
            await self.ok_done.acquire()
            await self.ok_done.acquire()
            await conn.wait_closed()
            self.owner.handle_message.assert_called_with(conn, Message.request('watchdog', mid=4))
        # Note: should only be one warning, not two
        self.assertEqual(1, len(cm.output))
        self.assertRegex(
            cm.output[0],
            r'^WARNING:aiokatcp\.connection:Connection closed .*: Connection lost \[.*\]$')
        # Allow the final watchdog to go through. This just provides test coverage
        # that Connection.write_message handles the writer having already gone away.
        self.ok_wait.release()
        await asynctest.exhaust_callbacks(self.loop)

    async def test_malformed(self) -> None:
        conn = self.make_connection(True)
        self.remote_writer.write(b'malformed\n')
        self.remote_writer.write_eof()
        with self.assertLogs('aiokatcp.connection', logging.WARN) as cm:
            # Wait for the close to go through
            await conn.wait_closed()
            self.owner.handle_message.assert_not_called()
        self.assertEqual(len(cm.output), 1)
        self.assertRegex(cm.output[0], 'Malformed message received.*')

    async def test_close_early(self) -> None:
        conn = self.make_connection(True)
        conn.close()
        await conn.wait_closed()

    async def test_close(self) -> None:
        conn = self.make_connection(True)
        await asynctest.exhaust_callbacks(self.loop)
        conn.close()
        await conn.wait_closed()

    async def test_exception(self) -> None:
        self.owner.handle_message.side_effect = RuntimeError('test error')
        conn = self.make_connection(True)
        await asynctest.exhaust_callbacks(self.loop)
        self.remote_writer.write(b'?watchdog[2]\n')
        self.remote_writer.close()
        with self.assertLogs('aiokatcp.connection', logging.ERROR) as cm:
            await conn.wait_closed()
        self.assertEqual(len(cm.output), 1)
        self.assertRegex(cm.output[0], '(?s)Exception in connection handler.*test error')



# Copyright 2017, 2019 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import inspect
import asyncio
import functools
import logging
import traceback
import io
import re
import time
from typing import (Set, Callable, Awaitable, Sequence, Iterable,
                    Optional, List, Any, TypeVar, cast)
# Only used in type comments, so flake8 complains
from typing import Dict    # noqa: F401

import aiokatcp
from . import core, connection, sensor
from .connection import FailReply, InvalidReply


logger = logging.getLogger(__name__)
_RequestReply = Awaitable[Optional[Sequence]]
_RequestHandler = Callable[['DeviceServer', 'RequestContext', core.Message], _RequestReply]
_T = TypeVar('_T')


class ClientConnection(connection.Connection):
    """Server's view of the connection from a single client."""

    def __init__(self, owner: 'DeviceServer',
                 reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        super().__init__(owner, reader, writer, True)
        #: Maps sensors to their samplers, for sensors that are being sampled
        self._samplers = {}     # type: Dict[sensor.Sensor, sensor.SensorSampler]

    def close(self) -> None:
        for sampler in self._samplers.values():
            sampler.close()
        self._samplers = {}
        super().close()

    def set_sampler(self, s: sensor.Sensor, sampler: Optional[sensor.SensorSampler]) -> None:
        """Set or clear the sampler for a sensor."""
        if s in self._samplers:
            self._samplers[s].close()
            del self._samplers[s]
        if sampler is not None:
            self._samplers[s] = sampler

    def get_sampler(self, s: sensor.Sensor) -> Optional[sensor.SensorSampler]:
        """Retrieve the sampler for a sensor"""
        return self._samplers.get(s)

    def sensor_update(self, s: sensor.Sensor, reading: sensor.Reading):
        """Report a new sensor value. This is used as the callback for the sampler."""
        msg = core.Message.inform(
            'sensor-status', reading.timestamp, 1, s.name, reading.status, reading.value)
        self.write_message(msg)


class RequestContext:
    """Interface for informs and replies to a request.

    Parameters
    ----------
    conn
        Client connection from which the request originated
    req
        The request itself
    """
    def __init__(self, conn: ClientConnection, req: core.Message) -> None:
        self.conn = conn
        self.req = req
        self._replied = False
        # Can't just wrap conn.logger, because LoggerAdapters can't be stacked
        self.logger = connection.ConnectionLoggerAdapter(
            logger, dict(address=conn.address, req=req))

    @property
    def replied(self) -> bool:
        """Whether a reply has currently been sent"""
        return self._replied

    def reply(self, *args: Any) -> None:
        """Send a reply to the request.

        Parameters
        ----------
        *args
            The fields of the reply

        Raises
        ------
        RuntimeError
            If the request has already been replied to.
        """
        if self._replied:
            raise RuntimeError('request ?{} has already been replied to'.format(self.req.name))
        msg = core.Message.reply_to_request(self.req, *args)
        self.conn.write_message(msg)
        self._replied = True

    def inform(self, *args: Any) -> None:
        """Send an inform in response to the request.

        Parameters
        ----------
        *args
            The fields of the inform

        Raises
        ------
        RuntimeError
            If the request has already been replied to.
        """
        if self._replied:
            raise RuntimeError('request ?{} has already been replied to'.format(self.req.name))
        msg = core.Message.inform_reply(self.req, *args)
        self.conn.write_message(msg)

    def informs(self, informs: Iterable[Iterable], *, send_reply=True) -> None:
        """Write a sequence of informs and send an ``ok`` reply with the count.

        Parameters
        ----------
        informs
            Each element is an iterable of fields in the inform.

        Raises
        ------
        RuntimeError
            If the request has already been replied to.
        """
        if self._replied:
            raise RuntimeError('request ?{} has already been replied to'.format(self.req.name))
        msgs = [core.Message.inform_reply(self.req, *inform) for inform in informs]
        if send_reply:
            msgs.append(core.Message.reply_to_request(self.req, core.Message.OK, len(msgs)))
        self.conn.write_messages(msgs)
        if send_reply:
            self._replied = True

    async def drain(self) -> None:
        """Wait for the outgoing queue to be below a threshold."""
        await self.conn.drain()


class DeviceServerMeta(type):
    @classmethod
    def _wrap_request(mcs, name: str, value: Callable[..., _RequestReply]) -> _RequestHandler:
        return connection.wrap_handler(name, value, 2)

    def __new__(mcs, name, bases, namespace, **kwds):
        namespace.setdefault('_request_handlers', {})
        for base in bases:
            namespace['_request_handlers'].update(getattr(base, '_request_handlers', {}))
        result = type.__new__(mcs, name, bases, namespace)
        request_handlers = getattr(result, '_request_handlers')
        for key, value in namespace.items():
            if key.startswith('request_') and inspect.isfunction(value):
                request_name = key[8:].replace('_', '-')
                if value.__doc__ is None:
                    raise TypeError('{} must have a docstring'.format(key))
                request_handlers[request_name] = mcs._wrap_request(request_name, value)
        return result


class DeviceServer(metaclass=DeviceServerMeta):
    """Server that handles katcp.

    Parameters
    ----------
    host
        Hostname to listen on (empty for all interfaces)
    port
        Port number to listen on
    limit
        Maximum line length in a request
    max_pending
        Maximum number of asynchronous requests that can be in progress.
        Once this number of reached, new requests are blocked.
    loop
        Event loop on which the server will run, defaulting to
        ``asyncio.get_event_loop()``.
    max_backlog
        Maximum backlog in the write queue to a client.

        If a message is to be sent to a client and it has more than this many
        bytes in its backlog, it is disconnected instead to prevent the server
        running out of memory. At present this is only applied to asynchronous
        informs, but it may be applied to other messages in future.

        If not specified it defaults to twice `limit`.
    """

    _request_handlers = {}   # type: Dict[str, _RequestHandler]

    VERSION = None           # type: str
    BUILD_STATE = None       # type: str

    class LogHandler(logging.Handler):
        """Log handler that issues log messages as ``#log`` informs.

        It is automatically initialised with a filter that discard log messages
        from the aiokatcp module, because otherwise one may get into an infinite
        recursion where log messages are communications with a client are
        communicated to the client.

        It is also initialised with a default formatter that emits only the
        the log message itself.

        Typical usage to inform clients about all log messages in the application
        is

        .. code:: python

            logging.getLogger().addHandler(DeviceServer.LogHandler(server))
        """
        def _self_filter(self, record: logging.LogRecord) -> bool:
            return not record.name.startswith('aiokatcp.') \
                and record.levelno >= self._server._log_level \
                and self._server._log_level != core.LogLevel.OFF

        def __init__(self, server: 'DeviceServer') -> None:
            super().__init__()
            self._server = server
            self.addFilter(self._self_filter)
            self.setFormatter(logging.Formatter('%(filename)s:%(lineno)d: %(message)s'))

        def emit(self, record: logging.LogRecord) -> None:
            try:
                msg = self.format(record)
                self._server.mass_inform(
                    'log', core.LogLevel.from_python(record.levelno),
                    record.created, record.name, msg)
            except Exception:
                self.handleError(record)

    def __init__(self, host: str, port: int, *,
                 limit: int = connection.DEFAULT_LIMIT, max_pending: int = 100,
                 max_backlog: int = None,
                 loop: asyncio.AbstractEventLoop = None) -> None:
        super().__init__()
        if not self.VERSION:
            raise TypeError('{.__name__} does not define VERSION'.format(self.__class__))
        if not self.BUILD_STATE:
            raise TypeError('{.__name__} does not define BUILD_STATE'.format(self.__class__))
        self._connections = set()  # type: Set[ClientConnection]
        self._pending = set()      # type: Set[asyncio.Task]
        self._pending_space = asyncio.Semaphore(value=max_pending, loop=loop)
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        self._limit = limit
        self.max_backlog = 2 * limit if max_backlog is None else max_backlog
        self._log_level = core.LogLevel.WARN
        self._server = None        # type: Optional[asyncio.events.AbstractServer]
        self._server_lock = asyncio.Lock(loop=loop)
        self._stopped = asyncio.Event(loop=loop)
        self._host = host
        self._port = port
        self._stopping = False
        self.sensors = sensor.SensorSet()
        self.sensors.add_remove_callback(
            functools.partial(self._remove_sensor_callback, self._connections))

    @staticmethod
    def _remove_sensor_callback(connections: Set[ClientConnection], sensor: sensor.Sensor):
        for conn in connections:
            conn.set_sampler(sensor, None)

    async def start(self) -> None:
        """Start the server running on the event loop.

        Raises
        ------
        RuntimeError
            if the server is already running
        """
        def factory():
            # Based on asyncio.start_server, but using ConvertCRProtocol
            reader = asyncio.StreamReader(limit=self._limit, loop=self.loop)
            protocol = connection.ConvertCRProtocol(
                reader, self._client_connected_cb, loop=self.loop)
            return protocol

        async with self._server_lock:
            if self._server is not None:
                raise RuntimeError('Server is already running')
            self._stopped.clear()
            self._server = await self.loop.create_server(factory, self._host, self._port)
            self._stopping = False

    async def on_stop(self) -> None:
        """Extension point for subclasses to run shutdown code.

        This is called after the TCP server has been shut down and all
        in-flight requests have been completed or cancelled. Subclasses
        should override this function rather than :meth:`stop` to run
        late shutdown code because this is called *before* the flag is
        set to wake up :meth:`join`.

        It is only called if the server was running when :meth:`stop` was
        called.
        """
        pass

    async def stop(self, cancel: bool = True) -> None:
        """Shut down the server.

        Parameters
        ----------
        cancel
            If true (default), cancel any pending asynchronous requests.
        """
        async with self._server_lock:
            self._stopping = True
            if self._server is not None:
                self._server.close()
                await self._server.wait_closed()
                self._server = None
                if self._pending:
                    for task in self._pending:
                        if cancel and not task.done():
                            task.cancel()
                    await asyncio.wait(list(self._pending), loop=self.loop)
                msg = core.Message.inform('disconnect', 'server shutting down')
                for client in list(self._connections):
                    client.write_message(msg)
                    client.close()
                    await client.wait_closed()
                await self.on_stop()
            self._stopped.set()

    def halt(self, cancel: bool = True) -> asyncio.Task:
        """Begin server shutdown, but do not wait for it to complete.

        Parameters
        ----------
        cancel
            If true (default), cancel any pending asynchronous requests.

        Returns
        -------
        task
            Task that is performing the stop.
        """
        return self.loop.create_task(self.stop(cancel))

    async def join(self) -> None:
        """Block until the server has stopped"""
        await self._stopped.wait()

    @property
    def server(self) -> Optional[asyncio.AbstractServer]:
        """Return the underlying TCP server"""
        return self._server

    def send_version_info(self, ctx: RequestContext, *, send_reply=True) -> None:
        """Send version information informs to the client.

        This is used for asynchronous #version-connect informs when a client
        connects and in response to a ?version-list request.

        Returns
        -------
        num_informs
            Number of informs sent
        """
        version = 'aiokatcp-{}'.format(aiokatcp.__version__)
        api_version = 'aiokatcp-{}'.format(aiokatcp.minor_version())
        ctx.informs([
            ('katcp-protocol', '5.0-MI'),
            ('katcp-library', api_version, version),
            ('katcp-device', self.VERSION, self.BUILD_STATE),
        ], send_reply=send_reply)

    def _write_async_message(self, conn: ClientConnection, msg: core.Message) -> None:
        conn.write_message(msg)
        if conn.writer is not None:
            transport = cast(asyncio.WriteTransport, conn.writer.transport)
            backlog = transport.get_write_buffer_size()
            if backlog > self.max_backlog:
                conn.logger.warning('Disconnecting client because it is too slow')
                transport.abort()
                conn.close()
                self._connections.discard(conn)

    async def _client_connected_cb(
            self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        async def cleanup():
            await conn.wait_closed()
            self._connections.discard(conn)
        conn = ClientConnection(self, reader, writer)
        # Copy the connection list, to avoid mutation while iterating and to
        # exclude the new connection from it.
        connections = list(self._connections)
        self._connections.add(conn)
        self.loop.create_task(cleanup())
        # Make a fake request context for send_version_info
        request = core.Message.request('version-connect')
        ctx = RequestContext(conn, request)
        self.send_version_info(ctx, send_reply=False)
        msg = core.Message.inform('client-connected', conn.address)
        for old_conn in connections:
            self._write_async_message(old_conn, msg)

    def _handle_request_done_callback(self, ctx: RequestContext, task: asyncio.Task) -> None:
        """Completion callback for request handlers.

        It deals with cancellation and error conditions, and ensures that
        exactly one reply is returned.
        """
        error_msg = None
        error_type = core.Message.FAIL
        if task in self._pending:
            self._pending.discard(task)
            self._pending_space.release()
        if task.cancelled():
            if ctx.replied:
                return     # Cancelled while draining the reply - not critical
            ctx.logger.info('request %r cancelled', ctx.req.name)
            error_msg = 'request cancelled'
        else:
            try:
                task.result()
            except FailReply as error:
                error_msg = str(error)
            except InvalidReply as error:
                error_msg = str(error)
                error_type = core.Message.INVALID
            except Exception:
                ctx.logger.exception('uncaught exception while handling %r',
                                     ctx.req.name, exc_info=True)
                output = io.StringIO('uncaught exception:\n')
                traceback.print_exc(file=output)
                error_msg = output.getvalue()
        if not ctx.replied:
            if error_msg is None:
                error_msg = 'request handler returned without replying'
            ctx.reply(error_type, error_msg)
        elif error_msg is not None:
            # We somehow replied before failing, so can't put the error
            # message in a reply - use an out-of-band inform instead.
            ctx.conn.write_message(core.Message.inform(
                'log', 'error', time.time(), __name__, error_msg))

    async def unhandled_request(self, ctx: RequestContext, req: core.Message) -> None:
        """Called when a request is received for which no handler is registered.

        Subclasses may override this to do dynamic handling.
        """
        raise InvalidReply('unknown request {}'.format(req.name))

    async def _handle_request(self, ctx: RequestContext) -> None:
        """Task for handling an incoming request.

        If the server is halted while the request is being handled, this task
        gets cancelled.
        """
        default = self.__class__.unhandled_request  # type: ignore
        handler = self._request_handlers.get(ctx.req.name, default)
        ret = await handler(self, ctx, ctx.req)
        if ctx.replied and ret is not None:
            raise RuntimeError('handler both replied and returned a value')
        if ret is None:
            ret = ()
        elif not isinstance(ret, tuple):
            ret = (ret,)
        if not ctx.replied:
            ctx.reply(core.Message.OK, *ret)
        await ctx.drain()

    async def handle_message(self, conn: ClientConnection, msg: core.Message) -> None:
        """Called by :class:`ClientConnection` for each incoming message."""
        if self._stopping:
            return
        if msg.mtype == core.Message.Type.REQUEST:
            await self._pending_space.acquire()
            ctx = RequestContext(conn, msg)
            task = self.loop.create_task(self._handle_request(ctx))
            self._pending.add(task)
            task.add_done_callback(functools.partial(self._handle_request_done_callback, ctx))
        else:
            pass
            # TODO: handle other message types

    def mass_inform(self, name: str, *args: Any) -> None:
        """Send an asynchronous inform to all clients.

        Parameters
        ----------
        name
            Inform name
        *args
            Fields for the inform
        """
        msg = core.Message.inform(name, *args)
        # Copy the connection list, because _write_async_message can mutate it.
        for conn in list(self._connections):
            self._write_async_message(conn, msg)

    async def request_help(self, ctx: RequestContext, name: str = None) -> None:
        """Return help on the available requests.

        Return a description of the available requests using a sequence of
        #help informs.

        Parameters
        ----------
        request : str, optional
            The name of the request to return help for (the default is to
            return help for all requests).

        Informs
        -------
        request : str
            The name of a request.
        description : str
            Documentation for the named request.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the help succeeded.
        informs : int
            Number of #help inform messages sent.

        Examples
        --------
        ::

            ?help
            #help halt ...description...
            #help help ...description...
            ...
            !help ok 5

            ?help halt
            #help halt ...description...
            !help ok 1

        """
        if name is None:
            # The cast is to keep mypy happy that __doc__ isn't None.
            informs = [(key, cast(str, handler.__doc__).splitlines()[0])
                       for key, handler in sorted(self._request_handlers.items())]
        else:
            try:
                handler = self._request_handlers[name]
            except KeyError as error:
                raise FailReply('request {} is not known'.format(name)) from error
            informs = [(name, cast(str, handler.__doc__))]
        ctx.informs(informs)

    async def request_halt(self, ctx: RequestContext) -> None:
        """Halt the device server.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether scheduling the halt succeeded.

        Examples
        --------
        ::

            ?halt
            !halt ok

        """
        self.halt()

    async def request_watchdog(self, ctx: RequestContext) -> None:
        """Check that the server is still alive.

        Returns
        -------
            success : {'ok'}

        Examples
        --------
        ::

            ?watchdog
            !watchdog ok

        """
        pass

    async def request_version_list(self, ctx: RequestContext) -> None:
        """Request the list of versions of roles and subcomponents.

        Informs
        -------
        name : str
            Name of the role or component.
        version : str
            A string identifying the version of the component. Individual
            components may define the structure of this argument as they
            choose. In the absence of other information clients should
            treat it as an opaque string.
        build_state_or_serial_number : str
            A unique identifier for a particular instance of a component.
            This should change whenever the component is replaced or updated.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the version list succeeded.
        informs : int
            Number of #version-list inform messages sent.

        Examples
        --------
        ::

            ?version-list
            #version-list katcp-protocol 5.0-MI
            #version-list katcp-library katcp-python-0.4 katcp-python-0.4.1-py2
            #version-list katcp-device foodevice-1.0 foodevice-1.0.0rc1
            !version-list ok 3

        """
        self.send_version_info(ctx)

    def _get_sensors(self, name: Optional[str]) -> List[sensor.Sensor]:
        """Retrieve the list of sensors, optionally filtered by a name or regex.

        Parameters
        ----------
        name
            Either a regex brackets by /'s, an exact name to match, or
            ``None`` to return all sensors.

        Raises
        ------
        FailReply
            if an invalid regex was provided, or an exact name was given
            which does not exist.
        """
        if name is None:
            matched = self.sensors.values()   # type: Iterable[sensor.Sensor]
        elif name.startswith('/') and name.endswith('/') and len(name) > 1:
            try:
                name_re = re.compile(name[1:-1])
            except re.error as error:
                raise FailReply(str(error)) from error
            matched = (sensor for sensor in self.sensors.values() if name_re.search(sensor.name))
        elif name not in self.sensors:
            # Do not change the wording: katcp.inspecting_client does a string
            # check for "Unknown sensor".
            raise FailReply('Unknown sensor {!r}'.format(name))
        else:
            matched = [self.sensors[name]]
        return sorted(matched, key=lambda sensor: sensor.name)

    async def request_sensor_list(self, ctx: RequestContext, name: str = None) -> int:
        r"""Request the list of sensors.

        The list of sensors is sent as a sequence of #sensor-list informs.

        Parameters
        ----------
        name : str, optional
            Name of the sensor to list (the default is to list all sensors).
            If name starts and ends with '/' it is treated as a regular
            expression and all sensors whose names contain the regular
            expression are returned.

        Informs
        -------
        name : str
            The name of the sensor being described.
        description : str
            Description of the named sensor.
        units : str
            Units for the value of the named sensor.
        type : str
            Type of the named sensor.
        params : list of str, optional
            Additional sensor parameters (type dependent). For discrete sensors
            the additional parameters are the allowed values. For all other
            types no additional parameters are sent.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the sensor list succeeded.
        informs : int
            Number of #sensor-list inform messages sent.

        Examples
        --------
        ::

            ?sensor-list
            #sensor-list psu.voltage PSU\_voltage. V float
            #sensor-list cpu.status CPU\_status. \@ discrete on off error
            ...
            !sensor-list ok 5

            ?sensor-list cpu.power.on
            #sensor-list cpu.power.on Whether\_CPU\_has\_power. \@ boolean
            !sensor-list ok 1

            ?sensor-list /voltage/
            #sensor-list psu.voltage PSU\_voltage. V float
            #sensor-list cpu.voltage CPU\_voltage. V float
            !sensor-list ok 2

        """
        sensors = self._get_sensors(name)
        for s in sensors:
            ctx.inform(s.name, s.description, s.units, s.type_name, *s.params)
        return len(sensors)

    async def request_sensor_value(self, ctx: RequestContext, name: str = None) -> None:
        """Request the value of a sensor or sensors.

        A list of sensor values as a sequence of #sensor-value informs.

        Parameters
        ----------
        name : str, optional
            Name of the sensor to poll (the default is to send values for all
            sensors). If name starts and ends with '/' it is treated as a
            regular expression and all sensors whose names contain the regular
            expression are returned.

        Informs
        -------
        timestamp : float
            Timestamp of the sensor reading in seconds since the Unix
            epoch, or milliseconds for katcp versions <= 4.
        count : {1}
            Number of sensors described in this #sensor-value inform. Will
            always be one. It exists to keep this inform compatible with
            #sensor-status.
        name : str
            Name of the sensor whose value is being reported.
        status : Sensor.Status
            Sensor status (see Sensor.Status enum)
        value : object
            Value of the named sensor. Type depends on the type of the sensor.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the list of values succeeded.
        informs : int
            Number of #sensor-value inform messages sent.

        Examples
        --------
        ::

            ?sensor-value
            #sensor-value 1244631611.415231 1 psu.voltage nominal 4.5
            #sensor-value 1244631611.415200 1 cpu.status warn off
            ...
            !sensor-value ok 5

            ?sensor-value cpu.power.on
            #sensor-value 1244631611.415231 1 cpu.power.on error 0
            !sensor-value ok 1

        """
        sensors = self._get_sensors(name)
        ctx.informs((s.timestamp, 1, s.name, s.status, s.value) for s in sensors)

    async def request_sensor_sampling(
            self, ctx: RequestContext, name: str,
            strategy: sensor.SensorSampler.Strategy = None,
            *args: bytes) -> tuple:
        """Configure or query the way a sensor is sampled.

        Sampled values are reported asynchronously using the #sensor-status
        message.

        Parameters
        ----------
        name
            Name of the sensor whose sampling strategy to query or configure.
        strategy
            Type of strategy to use to report the sensor value. The
            differential strategy type may only be used with integer or float
            sensors. If this parameter is supplied, it sets the new strategy.
        *args
            Additional strategy parameters (dependent on the strategy type).
            For the differential strategy, the parameter is an integer or float
            giving the amount by which the sensor value may change before an
            updated value is sent.
            For the period strategy, the parameter is the sampling period
            in float seconds.
            The event strategy has no parameters. Note that this has changed
            from KATCPv4.
            For the event-rate strategy, a minimum period between updates and
            a maximum period between updates (both in float seconds) must be
            given. If the event occurs more than once within the minimum period,
            only one update will occur. Whether or not the event occurs, the
            sensor value will be updated at least once per maximum period.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether the sensor-sampling request succeeded.
        name : str
            Name of the sensor queried or configured.
        strategy : :class:`.SensorSampler.Strategy`
            Name of the new or current sampling strategy for the sensor.
        params : list of str
            Additional strategy parameters (see description under Parameters).

        Examples
        --------
        ::

            ?sensor-sampling cpu.power.on
            !sensor-sampling ok cpu.power.on none

            ?sensor-sampling cpu.power.on period 500
            !sensor-sampling ok cpu.power.on period 500

        """
        try:
            s = self.sensors[name]
        except KeyError:
            raise FailReply('Unknown sensor {!r}'.format(name))
        if strategy is None:
            sampler = ctx.conn.get_sampler(s)
        else:
            try:
                observer = ctx.conn.sensor_update
                sampler = sensor.SensorSampler.factory(s, observer, self.loop, strategy, *args)
            except (TypeError, ValueError) as error:
                raise FailReply(str(error)) from error
            ctx.conn.set_sampler(s, sampler)
        if sampler is not None:
            params = sampler.parameters()
        else:
            params = (sensor.SensorSampler.Strategy.NONE,)
        return (name,) + params

    async def request_client_list(self, ctx: RequestContext) -> None:
        """Request the list of connected clients.

        The list of clients is sent as a sequence of #client-list informs.

        Informs
        -------
        addr : str
            The address of the client as host:port with host in dotted quad
            notation. If the address of the client could not be determined
            (because, for example, the client disconnected suddenly) then
            a unique string representing the client is sent instead.

        Returns
        -------
        success : {'ok', 'fail'}
            Whether sending the client list succeeded.
        informs : int
            Number of #client-list inform messages sent.

        Examples
        --------
        ::

            ?client-list
            #client-list 127.0.0.1:53600
            !client-list ok 1

        """
        ctx.informs((conn.address,) for conn in self._connections)

    async def request_log_level(
            self, ctx: RequestContext, level: core.LogLevel = None) -> core.LogLevel:
        """Query or set the current logging level.

        Parameters
        ----------
        level : {'all', 'trace', 'debug', 'info', 'warn', 'error', 'fatal', 'off'}, optional
            Name of the logging level to set the device server to (the default
            is to leave the log level unchanged).

        Returns
        -------
        success : {'ok', 'fail'}
            Whether the request succeeded.
        level : {'all', 'trace', 'debug', 'info', 'warn', 'error', 'fatal', 'off'}
            The log level after processing the request.

        Examples
        --------
        ::

            ?log-level
            !log-level ok warn

            ?log-level info
            !log-level ok info

        """
        if level is not None:
            self._log_level = level
        return self._log_level



# Copyright 2017, 2019 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import enum
import re
import asyncio
import ipaddress
import unittest
import unittest.mock
import logging
from typing import Tuple, Iterable, Union, Pattern, Type
from typing import List   # noqa: F401

import asynctest

import aiokatcp
from aiokatcp.core import Message, Address
from aiokatcp.connection import FailReply
from aiokatcp.server import DeviceServer, RequestContext
from aiokatcp.sensor import Sensor, SensorSampler
from .test_utils import timelimit


class Foo(enum.Enum):
    FIRST_VALUE = 1
    SECOND_VALUE = 2


class DummyServer(DeviceServer):
    VERSION = 'dummy-1.0'
    BUILD_STATE = 'dummy-build-1.0.0'

    def __init__(self, *, loop=None):
        super().__init__('127.0.0.1', 0, loop=loop)
        self.event = asyncio.Event(loop=self.loop)
        self.wait_reached = asyncio.Event(loop=self.loop)
        sensor = Sensor(int, 'counter-queries', 'number of ?counter queries',
                        default=0,
                        initial_status=Sensor.Status.NOMINAL)
        self.sensors.add(sensor)
        self._counter = sensor
        sensor = Sensor(Foo, 'foo', 'nonsense')
        self.sensors.add(sensor)
        sensor = Sensor(float, 'float-sensor', 'generic float sensor')
        self.sensors.add(sensor)
        sensor = Sensor(int, 'auto-override', 'overrides the auto strategy',
                        auto_strategy=SensorSampler.Strategy.PERIOD,
                        auto_strategy_parameters=(2.5,))
        self.sensors.add(sensor)
        self.on_stop_called = 0

    async def request_increment_counter(self, ctx: RequestContext) -> None:
        """Increment a counter"""
        self._counter.value += 1

    async def request_echo(self, ctx: RequestContext, *args: str) -> Tuple:
        """Return the arguments to the caller"""
        return tuple(args)

    async def request_bytes_arg(self, ctx: RequestContext, some_bytes) -> bytes:
        """An argument with no annotation"""
        if not isinstance(some_bytes, bytes):
            raise FailReply('expected bytes, got {}'.format(type(some_bytes)))
        return some_bytes

    async def request_double(self, ctx: RequestContext, number: float, scale: float = 2) -> float:
        """Take a float and double it"""
        return scale * number

    async def request_wait(self, ctx: RequestContext) -> None:
        """Wait for an internal event to fire"""
        self.wait_reached.set()
        await self.event.wait()

    async def request_crash(self, ctx: RequestContext) -> None:
        """Request that always raises an exception"""
        raise RuntimeError("help I fell over")

    async def on_stop(self) -> None:
        self.on_stop_called += 1


class BadServer(DummyServer):
    """Server with some badly-behaved requests."""
    async def request_double_reply(self, ctx: RequestContext) -> None:
        """Tries to send a reply twice"""
        ctx.reply('ok', 'reply1')
        ctx.reply('ok', 'reply2')

    async def request_reply_return(self, ctx: RequestContext) -> int:
        """Sends an explicit reply and returns a value"""
        ctx.reply('ok', 'reply1')
        return 3

    async def request_inform_after_reply(self, ctx: RequestContext) -> None:
        """Call inform after reply"""
        ctx.reply('ok', 'reply1')
        ctx.inform('inform1')

    async def request_informs_after_reply(self, ctx: RequestContext) -> None:
        """Call informs after reply"""
        ctx.reply('ok', 'reply1')
        ctx.informs(['inform1'])


class DeviceServerTestMixin(asynctest.TestCase):
    """Mixin for device server tests.

    It creates a server and a reader and writer that form a client.
    """
    async def make_server(self, server_cls: Type[DummyServer]) -> DummyServer:
        server = server_cls(loop=self.loop)
        await server.start()
        self.addCleanup(server.stop)
        return server

    async def make_client(self, server) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        host, port = server.server.sockets[0].getsockname()    # type: ignore
        remote_reader, remote_writer = await asyncio.open_connection(
            host, port, loop=self.loop)
        self.addCleanup(remote_writer.close)
        return remote_reader, remote_writer

    async def setUp(self) -> None:
        self.server = await self.make_server(DummyServer)
        self.remote_reader, self.remote_writer = await self.make_client(self.server)

    async def _readline(self) -> bytes:
        return await self.remote_reader.readline()

    async def _write(self, data: bytes) -> None:
        self.remote_writer.write(data)
        await self.remote_writer.drain()

    async def _check_reply(self, lines: Iterable[Union[bytes, Pattern[bytes]]]) -> None:
        for expected in lines:
            actual = await self._readline()
            if isinstance(expected, bytes):
                self.assertEqual(ascii(actual), ascii(expected))
            else:
                self.assertRegex(actual, expected)

    async def get_version_info(self, prefix: bytes = b'#version-connect') -> None:
        await self._check_reply([
            prefix + b' katcp-protocol 5.0-MI\n',
            prefix + ' katcp-library aiokatcp-{} aiokatcp-{}\n'.format(
                aiokatcp.minor_version(), aiokatcp.__version__).encode('ascii'),
            prefix + b' katcp-device dummy-1.0 dummy-build-1.0.0\n'])


@timelimit
class TestDeviceServer(DeviceServerTestMixin, asynctest.TestCase):
    async def setUp(self) -> None:
        patcher = unittest.mock.patch('time.time', return_value=123456789.0)
        patcher.start()
        self.addCleanup(patcher.stop)
        await super().setUp()

    async def test_start_twice(self) -> None:
        """Calling start twice raises :exc:`RuntimeError`"""
        with self.assertRaises(RuntimeError):
            await self.server.start()

    async def test_carriage_return(self) -> None:
        await self.get_version_info()
        await self._write(b'?watchdog[2]\r')
        await self._check_reply([b'!watchdog[2] ok\n'])

    async def test_help(self) -> None:
        await self.get_version_info()
        await self._write(b'?help[1]\n')
        commands = sorted([
            'increment-counter', 'echo', 'bytes-arg', 'double', 'wait', 'crash',
            'client-list', 'log-level', 'sensor-list', 'sensor-sampling',
            'sensor-value', 'halt', 'help', 'watchdog', 'version-list'
        ])
        expected = [re.compile(br'^#help\[1\] ' + cmd.encode('ascii') + b' [^ ]+$\n')
                    for cmd in commands]   # type: List[Union[bytes, Pattern[bytes]]]
        expected.append(b'!help[1] ok %d\n' % (len(commands),))
        await self._check_reply(expected)

    async def test_help_command(self) -> None:
        await self.get_version_info()
        await self._write(b'?help[1] increment-counter\n')
        await self._check_reply([
            br'#help[1] increment-counter Increment\_a\_counter' + b'\n',
            b'!help[1] ok 1\n'])

    async def test_help_bad_command(self) -> None:
        await self.get_version_info()
        await self._write(b'?help[1] increment_counter\n')  # note: _ instead of -
        await self._check_reply([
            br'!help[1] fail request\_increment_counter\_is\_not\_known' + b'\n'])

    async def test_watchdog(self) -> None:
        await self.get_version_info()
        await self._write(b'?watchdog[2]\n')
        await self._check_reply([b'!watchdog[2] ok\n'])

    async def test_client_list(self) -> None:
        await self.get_version_info()
        await self._write(b'?client-list[3]\n')
        host, port = self.remote_writer.get_extra_info('sockname')
        client_addr = Address(ipaddress.ip_address(host), port)
        await self._check_reply([
            b'#client-list[3] ' + bytes(client_addr) + b'\n',
            b'!client-list[3] ok 1\n'])

    async def test_sensor_list_no_filter(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-list[4]\n')
        await self._check_reply([
            br'#sensor-list[4] auto-override overrides\_the\_auto\_strategy \@ integer' + b'\n',
            br'#sensor-list[4] counter-queries number\_of\_?counter\_queries \@ integer' + b'\n',
            br'#sensor-list[4] float-sensor generic\_float\_sensor \@ float' + b'\n',
            br'#sensor-list[4] foo nonsense \@ discrete first-value second-value' + b'\n',
            b'!sensor-list[4] ok 4\n'])

    async def test_sensor_list_simple_filter(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-list[5] foo\n')
        await self._check_reply([
            br'#sensor-list[5] foo nonsense \@ discrete first-value second-value' + b'\n',
            b'!sensor-list[5] ok 1\n'])

    async def test_sensor_list_simple_no_match(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-list[5] bar\n')
        await self._check_reply([
            br"!sensor-list[5] fail Unknown\_sensor\_'bar'" + b'\n'])

    async def test_sensor_list_regex_filter(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-list[6] /^[b-z][a-z-]+/\n')
        await self._check_reply([
            br'#sensor-list[6] counter-queries number\_of\_?counter\_queries \@ integer' + b'\n',
            br'#sensor-list[6] float-sensor generic\_float\_sensor \@ float' + b'\n',
            br'#sensor-list[6] foo nonsense \@ discrete first-value second-value' + b'\n',
            b'!sensor-list[6] ok 3\n'])

    async def test_sensor_list_regex_filter_empty(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-list[7] /unmatched/\n')
        await self._check_reply([
            b'!sensor-list[7] ok 0\n'])

    async def test_sensor_list_regex_bad_regex(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-list[8] /(/\n')
        await self._check_reply([
            re.compile(br'^!sensor-list\[8\] fail .+\n$')])

    async def test_sensor_value(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-value[9]\n')
        await self._check_reply([
            b'#sensor-value[9] 123456789.0 1 auto-override unknown 0\n',
            b'#sensor-value[9] 123456789.0 1 counter-queries nominal 0\n',
            b'#sensor-value[9] 123456789.0 1 float-sensor unknown 0.0\n',
            b'#sensor-value[9] 123456789.0 1 foo unknown first-value\n',
            b'!sensor-value[9] ok 4\n'])

    async def test_sensor_value_filter(self) -> None:
        # The other filter tests are omitted since they're covered by the
        # sensor-list tests.
        await self.get_version_info()
        await self._write(b'?sensor-value[9] counter-queries\n')
        await self._check_reply([
            b'#sensor-value[9] 123456789.0 1 counter-queries nominal 0\n',
            b'!sensor-value[9] ok 1\n'])

    async def test_version_list(self) -> None:
        await self.get_version_info()
        await self._write(b'?version-list[10]\n')
        await self.get_version_info(prefix=b'#version-list[10]')
        await self._check_reply([b'!version-list[10] ok 3\n'])

    async def test_halt(self) -> None:
        await self.get_version_info()
        await self._write(b'?halt[11]\n')
        await self._check_reply([
            b'!halt[11] ok\n',
            br'#disconnect server\_shutting\_down' + b'\n',
            b''])   # Empty string indicates EOF
        await self.server.join()

    async def test_halt_while_waiting(self) -> None:
        await self.get_version_info()
        await self._write(b'?wait[1]\n?halt[2]\n')
        await self._check_reply([
            b'!halt[2] ok\n',
            b'!wait[1] fail request\\_cancelled\n',
            b'#disconnect server\\_shutting\\_down\n',
            b''])    # Empty line indicates EOF

    async def test_too_few_params(self) -> None:
        await self.get_version_info()
        await self._write(b'?double\n')
        await self._check_reply([re.compile(br'^!double fail .*\n$')])

    async def test_too_many_params(self) -> None:
        await self.get_version_info()
        await self._write(b'?double 1 2 3\n')
        await self._check_reply([re.compile(br'^!double fail .*\n$')])

    async def test_unknown_request(self) -> None:
        await self.get_version_info()
        await self._write(b'?cheese[1]\n')
        await self._check_reply([
            re.compile(br'^!cheese\[1\] invalid unknown\\_request\\_cheese\n$')])

    async def test_crash_request(self) -> None:
        await self.get_version_info()
        await self._write(b'?crash\n')
        await self._check_reply([
            re.compile(br'^!crash fail .*help\\_I\\_fell\\_over.*$')])

    async def test_variadic(self) -> None:
        """Test a request that takes a ``*args``."""
        await self.get_version_info()
        await self._write(b'?echo hello world\n')
        await self._check_reply([b'!echo ok hello world\n'])

    async def test_no_annotation(self) -> None:
        """Argument without annotation is passed through raw"""
        await self.get_version_info()
        await self._write(b'?bytes-arg raw\n')
        await self._check_reply([b'!bytes-arg ok raw\n'])

    async def test_bad_arg_type(self) -> None:
        await self.get_version_info()
        await self._write(b'?double bad\n')
        await self._check_reply([
            br"!double fail could\_not\_convert\_string\_to\_float:\_'bad'" + b'\n'])

    async def test_concurrent(self) -> None:
        await self.get_version_info()
        await self._write(b'?wait[1]\n?echo[2] test\n')
        await self._check_reply([b'!echo[2] ok test\n'])
        self.server.event.set()
        await self._check_reply([b'!wait[1] ok\n'])

    async def test_client_connected_inform(self) -> None:
        """A second client connecting sends ``#client-connected`` to the first"""
        await self.get_version_info()
        host, port = self.server.server.sockets[0].getsockname()    # type: ignore
        reader2, writer2 = await asyncio.open_connection(host, port, loop=self.loop)
        self.addCleanup(writer2.close)
        client_host, client_port = writer2.get_extra_info('sockname')
        client_addr = Address(ipaddress.ip_address(client_host), client_port)
        client_addr_bytes = bytes(client_addr)
        await self._check_reply([b'#client-connected ' + client_addr_bytes + b'\n'])

    async def test_message_while_stopping(self) -> None:
        await self.get_version_info()
        await self._write(b'?wait\n')
        # Wait for the request_wait to be launched
        await self.server.wait_reached.wait()
        # Start stopping the server, but wait for outstanding tasks
        stop_task = self.loop.create_task(self.server.stop(cancel=False))
        await self._write(b'?watchdog\n')   # Should be ignored, because we're stopping
        # Give the ?watchdog time to make it through to the message handler
        await asyncio.sleep(0.1, loop=self.loop)
        self.server.event.set()   # Releases the ?wait
        await self._check_reply([
            b'!wait ok\n',
            b'#disconnect server\\_shutting\\_down\n',
            b''])
        await stop_task

    async def test_mass_inform(self) -> None:
        await self.get_version_info()
        self.server.mass_inform('test-inform', 123)
        await self._check_reply([b'#test-inform 123\n'])

    async def test_log_level(self) -> None:
        # We have to use a name that doesn't start with aiokatcp., because
        # those get filtered out.
        logger = logging.getLogger('_aiokatcp.test.test_logger')
        logger.propagate = False
        logger.level = logging.DEBUG
        handler = DummyServer.LogHandler(self.server)
        logger.addHandler(handler)
        self.addCleanup(logger.removeHandler, handler)
        await self.get_version_info()
        await self._write(b'?log-level\n')
        await self._check_reply([b'!log-level ok warn\n'])
        logger.info('foo')  # Should not be reported
        await self._write(b'?log-level info\n')
        await self._check_reply([b'!log-level ok info\n'])
        logger.info('bar')  # Should be reported
        await self._check_reply([
            re.compile(br'^#log info 123456789\.0 _aiokatcp.test.test_logger '
                       br'test_server\.py:\d+:\\_bar\n')])

    async def test_on_stop(self) -> None:
        self.assertEqual(self.server.on_stop_called, 0)
        self.server.halt()
        await self.server.join()
        self.assertEqual(self.server.on_stop_called, 1)
        # on_stop should not be called when already stopped
        await self.server.stop()
        self.assertEqual(self.server.on_stop_called, 1)

    async def test_slow_client(self) -> None:
        self.server.max_backlog = 32768
        # We need to be sure to push in lots of data, because some of it will
        # be absorbed by TCP socket buffers, receive buffers etc.
        big_str = b'x' * 200000
        self.assertEqual(len(self.server._connections), 1)
        for i in range(100):
            self.server.mass_inform('big', big_str)
        self.assertEqual(len(self.server._connections), 0)


class TestDeviceServerClocked(DeviceServerTestMixin, asynctest.ClockedTestCase):
    """Tests for :class:`.DeviceServer` that use a fake clock.

    These are largely to do with sensor sampling.
    """
    def _mock_time(self) -> float:
        return self.loop.time() - self._start_time + 123456789.0

    async def setUp(self) -> None:
        self._start_time = self.loop.time()
        patcher = unittest.mock.patch('time.time', new=self._mock_time)
        patcher.start()
        self.addCleanup(patcher.stop)
        await super().setUp()

    async def test_sensor_sampling_invalid(self) -> None:
        """Invalid strategy for ``?sensor-strategy``"""
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor carrot\n')
        await self._check_reply([
            br"!sensor-sampling fail b'carrot'\_is\_not\_a\_valid\_value\_for\_Strategy" + b'\n'])

    async def test_sensor_sampling_none_params(self) -> None:
        """None strategy must not accept parameters"""
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor none 4\n')
        await self._check_reply([
            br'!sensor-sampling fail expected\_0\_strategy\_arguments,\_found\_1' + b'\n'])

    async def test_sensor_sampling_too_few_params(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor event-rate 4\n')
        await self._check_reply([
            br'!sensor-sampling fail expected\_2\_strategy\_arguments,\_found\_1' + b'\n'])

    async def test_sensor_sampling_too_many_params(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor event-rate 4 5 6\n')
        await self._check_reply([
            br'!sensor-sampling fail expected\_2\_strategy\_arguments,\_found\_3' + b'\n'])

    async def test_sensor_sampling_bad_parameter(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor period foo\n')
        await self._check_reply([
            br"!sensor-sampling fail could\_not\_convert\_string\_to\_float:\_'foo'" + b'\n'])

    async def test_sensor_sampling_auto(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor auto\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 float-sensor unknown 0.0\n',
            b'!sensor-sampling ok float-sensor auto\n'])
        # Set to a new value
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 1.25
        # Set to the same value again - must still update
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 1.25
        await self._check_reply([
            b'#sensor-status 123456790.0 1 float-sensor nominal 1.25\n',
            b'#sensor-status 123456791.0 1 float-sensor nominal 1.25\n'])

    async def test_sensor_sampling_period(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor period 2.5\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 float-sensor unknown 0.0\n',
            b'!sensor-sampling ok float-sensor period 2.5\n'])
        await self.advance(3.0)
        self.server.sensors['float-sensor'].value = 1.25
        await self.advance(5.0)
        await self._write(b'?watchdog\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 float-sensor unknown 0.0\n',
            b'#sensor-status 123456792.0 1 float-sensor nominal 1.25\n',
            b'#sensor-status 123456792.0 1 float-sensor nominal 1.25\n',
            b'!watchdog ok\n'])

    async def test_sensor_sampling_period_zero(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor period 0.0\n')
        await self._check_reply([
            br'!sensor-sampling fail period\_must\_be\_positive' + b'\n'])

    async def test_sensor_sampling_event(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor event\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 float-sensor unknown 0.0\n',
            b'!sensor-sampling ok float-sensor event\n'])
        # Set to a new value
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 1.25
        # Set to the same value again - must not update
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 1.25
        # Set to the same value, but change the status
        await self.advance(1.0)
        self.server.sensors['float-sensor'].set_value(1.25, status=Sensor.Status.WARN)
        await self._check_reply([
            b'#sensor-status 123456790.0 1 float-sensor nominal 1.25\n',
            b'#sensor-status 123456792.0 1 float-sensor warn 1.25\n'])

    async def test_sensor_sampling_differential_bad_type(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling foo differential 1\n')
        await self._check_reply([
            br'!sensor-sampling fail differential\_strategies\_only\_valid\_for\_integer'
            br'\_and\_float\_sensors' + b'\n'])

    async def test_sensor_sampling_differential(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor differential 1.5\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 float-sensor unknown 0.0\n',
            b'!sensor-sampling ok float-sensor differential 1.5\n'])
        # Set to a valid value, less than the difference away (but a status change)
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 1.25
        # Set to the same value
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 1.25
        # Set to a new value, within the delta
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 2.5
        # Set to a new value, within the delta of the last update
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 0.2
        # Set to a new value, outside the delta
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = -1.0
        await self._check_reply([
            b'#sensor-status 123456790.0 1 float-sensor nominal 1.25\n',
            b'#sensor-status 123456794.0 1 float-sensor nominal -1.0\n'])

    async def test_sensor_sampling_event_rate(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor event-rate 1.0 10.0\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 float-sensor unknown 0.0\n',
            b'!sensor-sampling ok float-sensor event-rate 1.0 10.0\n'])
        # Set to a valid value, after the short period has passed
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 1.25
        # Change the twice value, within the short period
        await self.advance(0.25)
        self.server.sensors['float-sensor'].value = 1.0
        await self.advance(0.25)
        self.server.sensors['float-sensor'].value = 0.5
        # Leave it alone until just before the long period triggers
        await self.advance(10.25)
        # Inject a message directly into the server, so that its reply can
        # be reliably sequenced relative to the other informs.
        conn = next(iter(self.server._connections))
        await self.server.handle_message(conn, Message.request('watchdog'))
        # Wait for the long period to trigger
        await self.advance(0.5)

        await self._check_reply([
            b'#sensor-status 123456790.0 1 float-sensor nominal 1.25\n',
            b'#sensor-status 123456790.5 1 float-sensor nominal 0.5\n',
            b'!watchdog ok\n',
            b'#sensor-status 123456790.5 1 float-sensor nominal 0.5\n'])

    async def test_sensor_sampling_differential_rate(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor differential-rate 1.5 1.0 10.0\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 float-sensor unknown 0.0\n',
            b'!sensor-sampling ok float-sensor differential-rate 1.5 1.0 10.0\n'])
        # Set to a valid value, after the short period has passed
        await self.advance(1.0)
        self.server.sensors['float-sensor'].value = 1.25
        # Change the twice value, within the short period
        await self.advance(0.25)
        self.server.sensors['float-sensor'].value = 5.0
        await self.advance(0.25)
        self.server.sensors['float-sensor'].value = 0.5
        # Let it trigger, then change it, but not enough to trigger
        await self.advance(4.0)
        self.server.sensors['float-sensor'].value = 1.5
        # Leave it alone until just before the long period triggers
        await self.advance(6.25)
        # Inject a message directly into the server, so that its reply can
        # be reliably sequenced relative to the other informs.
        conn = next(iter(self.server._connections))
        await self.server.handle_message(conn, Message.request('watchdog'))
        # Wait for the long period to trigger
        await self.advance(0.5)

        await self._check_reply([
            b'#sensor-status 123456790.0 1 float-sensor nominal 1.25\n',
            b'#sensor-status 123456790.5 1 float-sensor nominal 0.5\n',
            b'!watchdog ok\n',
            b'#sensor-status 123456794.5 1 float-sensor nominal 1.5\n'])

    async def test_sensor_sampling_none(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor event\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 float-sensor unknown 0.0\n',
            b'!sensor-sampling ok float-sensor event\n'])
        await self._write(b'?sensor-sampling float-sensor none\n')
        await self._check_reply([
            b'!sensor-sampling ok float-sensor none\n'])
        self.server.sensors['float-sensor'].value = 1.0
        # Must not report the update
        await self._write(b'?watchdog\n')
        await self._check_reply([b'!watchdog ok\n'])

    async def test_sensor_sampling_query(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling float-sensor event\n'
                          b'?sensor-sampling float-sensor\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 float-sensor unknown 0.0\n',
            b'!sensor-sampling ok float-sensor event\n',
            b'!sensor-sampling ok float-sensor event\n'])

    async def test_sensor_sampling_unknown_sensor(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling bad-sensor event\n')
        await self._check_reply([
            br"!sensor-sampling fail Unknown\_sensor\_'bad-sensor'" + b'\n'])

    async def test_sensor_sampling_auto_override(self) -> None:
        await self.get_version_info()
        await self._write(b'?sensor-sampling auto-override auto\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 auto-override unknown 0\n',
            b'!sensor-sampling ok auto-override auto\n'])
        await self.advance(3.0)
        self.server.sensors['auto-override'].value = 1
        await self.advance(5.0)
        await self._write(b'?watchdog\n')
        await self._check_reply([
            b'#sensor-status 123456789.0 1 auto-override unknown 0\n',
            b'#sensor-status 123456792.0 1 auto-override nominal 1\n',
            b'#sensor-status 123456792.0 1 auto-override nominal 1\n',
            b'!watchdog ok\n'])


@timelimit
class TestBadDeviceServer(DeviceServerTestMixin, asynctest.TestCase):
    async def setUp(self) -> None:
        patcher = unittest.mock.patch('time.time', return_value=123456789.0)
        patcher.start()
        self.addCleanup(patcher.stop)
        self.server = await self.make_server(BadServer)
        self.remote_reader, self.remote_writer = await self.make_client(self.server)

    async def test_double_reply(self) -> None:
        await self.get_version_info()
        await self._write(b'?double-reply\n')
        await self._check_reply([
            b'!double-reply ok reply1\n',
            re.compile(br'^#log error [0-9.]+ aiokatcp\.server Traceback')])

    async def test_reply_return(self) -> None:
        await self.get_version_info()
        await self._write(b'?reply-return\n')
        await self._check_reply([
            b'!reply-return ok reply1\n',
            re.compile(br'^#log error [0-9.]+ aiokatcp\.server Traceback')])

    async def test_inform_after_reply(self) -> None:
        await self.get_version_info()
        await self._write(b'?inform-after-reply\n')
        await self._check_reply([
            b'!inform-after-reply ok reply1\n',
            re.compile(br'^#log error [0-9.]+ aiokatcp\.server Traceback')])

    async def test_informs_after_reply(self) -> None:
        await self.get_version_info()
        await self._write(b'?informs-after-reply\n')
        await self._check_reply([
            b'!informs-after-reply ok reply1\n',
            re.compile(br'^#log error [0-9.]+ aiokatcp\.server Traceback')])


class TestDeviceServerMeta(unittest.TestCase):
    """Test that the metaclass picks up invalid constructions"""
    def test_missing_help(self) -> None:
        with self.assertRaises(TypeError):
            class MyBadServer(DummyServer):
                def request_no_help(self):
                    pass

    def test_too_few_parameters(self) -> None:
        with self.assertRaises(TypeError):
            class MyBadServer(DummyServer):
                def request_too_few(self):
                    """Not enough parameters"""

    def test_missing_version(self) -> None:
        with self.assertRaises(TypeError) as cm:
            class MyBadServer(DeviceServer):
                BUILD_INFO = 'build-info'
            MyBadServer('127.0.0.1', 0)
        self.assertEqual(str(cm.exception), 'MyBadServer does not define VERSION')

    def test_missing_build_state(self) -> None:
        with self.assertRaises(TypeError) as cm:
            class MyBadServer(DeviceServer):
                VERSION = 'version'
            MyBadServer('127.0.0.1', 0)
        self.assertEqual(str(cm.exception), 'MyBadServer does not define BUILD_STATE')



# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import asyncio
import re
import ipaddress
import socket
import time
import inspect
import functools
import types
from typing import Any, Optional, Iterable, Callable, cast

import decorator

from . import core


logger = logging.getLogger(__name__)
DEFAULT_LIMIT = 16 * 1024**2
_BLANK_RE = re.compile(br'^[ \t]*[\r\n]?$')


class ConvertCRProtocol(asyncio.StreamReaderProtocol):
    """Protocol that converts incoming carriage returns to newlines.

    This simplifies extracting the data with :class:`asyncio.StreamReader`,
    whose :meth:`~asyncio.StreamReader.readuntil` method is limited to a single
    separator.
    """
    def data_received(self, data: bytes) -> None:
        super().data_received(data.replace(b'\r', b'\n'))


async def _discard_to_eol(stream: asyncio.StreamReader) -> None:
    """Discard all data up to and including the next newline, or end of file."""
    while True:
        try:
            await stream.readuntil()
        except asyncio.IncompleteReadError:
            break     # EOF reached
        except asyncio.LimitOverrunError as error:
            # Extract the data that's already in the buffer
            consumed = error.consumed
            await stream.readexactly(consumed)
        else:
            break


async def read_message(stream: asyncio.StreamReader) -> Optional[core.Message]:
    """Read a single message from an asynchronous stream.

    If EOF is reached before reading the newline, returns ``None`` if
    there was no data, otherwise raises
    :exc:`aiokatcp.core.KatcpSyntaxError`.

    Parameters
    ----------
    stream
        Input stream

    Raises
    ------
    aiokatcp.core.KatcpSyntaxError
        if the line was too long or malformed.
    """
    while True:
        try:
            raw = await stream.readuntil()
        except asyncio.IncompleteReadError as error:
            raw = error.partial
            if not raw:
                return None    # End of stream reached
        except asyncio.LimitOverrunError:
            await _discard_to_eol(stream)
            raise core.KatcpSyntaxError('Message exceeded stream buffer size')
        if not _BLANK_RE.match(raw):
            return core.Message.parse(raw)


class FailReply(Exception):
    """Indicate to the remote end that a request failed, without backtrace"""


class InvalidReply(Exception):
    """Indicate to the remote end that a request was unrecognised"""


class ConnectionLoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        return '{} [{}]'.format(msg, self.extra['address']), kwargs


class Connection:
    def __init__(self, owner: Any,
                 reader: asyncio.StreamReader, writer: asyncio.StreamWriter,
                 is_server: bool) -> None:
        # Set TCP_NODELAY to avoid unnecessary transmission delays. This is
        # on by default in asyncio from Python 3.6, but not in 3.5.
        writer.get_extra_info('socket').setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.owner = owner
        self.reader = reader
        self.writer = writer  # type: Optional[asyncio.StreamWriter]
        host, port, *_ = writer.get_extra_info('peername')
        self.address = core.Address(ipaddress.ip_address(host), port)
        self._drain_lock = asyncio.Lock(loop=owner.loop)
        self.is_server = is_server
        self.logger = ConnectionLoggerAdapter(logger, dict(address=self.address))
        self._task = self.owner.loop.create_task(self._run())
        self._task.add_done_callback(self._done_callback)
        self._closing = False
        self._closed_event = asyncio.Event(loop=owner.loop)

    def _close_writer(self):
        if self.writer is not None:
            self.writer.close()
            self.writer = None

    def write_messages(self, msgs: Iterable[core.Message]) -> None:
        """Write a stream of messages to the connection.

        Connection errors are logged and swallowed.
        """
        if self.writer is None:
            return     # We previously detected that it was closed
        try:
            # Normally this would be checked by the internals of
            # self.writer.drain and bubble out to self.drain, but there is no
            # guaranteed that self.drain will be called in the near future
            # (see Github issue #11).
            if self.writer.transport.is_closing():
                raise ConnectionResetError('Connection lost')
            raw = b''.join(bytes(msg) for msg in msgs)
            self.writer.write(raw)
            self.logger.debug('Sent message %r', raw)
        except ConnectionError as error:
            self.logger.warning('Connection closed before message could be sent: %s', error)
            self._close_writer()

    def write_message(self, msg: core.Message) -> None:
        """Write a message to the connection.

        Connection errors are logged and swallowed.
        """
        self.write_messages([msg])

    async def drain(self) -> None:
        """Block until the outgoing write buffer is small enough."""
        # The Python 3.5 implementation of StreamWriter.drain is not reentrant,
        # so we use a lock.
        async with self._drain_lock:
            if self.writer is not None:
                try:
                    await self.writer.drain()
                except ConnectionError as error:
                    self.logger.warning('Connection closed while draining: %s', error)
                    self._close_writer()

    async def _run(self) -> None:
        while True:
            # If the output buffer gets too full, pause processing requests
            await self.drain()
            try:
                msg = await read_message(self.reader)
            except core.KatcpSyntaxError as error:
                self.logger.warning('Malformed message received', exc_info=True)
                if self.is_server:
                    # TODO: #log informs are supposed to go to all clients
                    self.write_message(
                        core.Message.inform('log', 'error', time.time(), __name__, str(error)))
            except ConnectionResetError:
                # Client closed connection without consuming everything we sent it.
                break
            else:
                if msg is None:   # EOF received
                    break
                self.logger.debug('Received message %r', bytes(msg))
                await self.owner.handle_message(self, msg)

    def _done_callback(self, task: asyncio.Future) -> None:
        self._closed_event.set()
        if not task.cancelled():
            try:
                task.result()
            except Exception:
                self.logger.exception('Exception in connection handler')

    def close(self) -> None:
        """Start closing the connection.

        Any currently running message handler will be cancelled. The closing
        process completes asynchronously. Use :meth:`wait_closed` to wait for
        things to be completely closed off.
        """
        if not self._closing:
            self._task.cancel()
            self._close_writer()
            self._closing = True

    async def wait_closed(self) -> None:
        """Wait until the connection is closed.

        This can be used either after :meth:`close`, or without :meth:`close`
        to wait for the remote end to close the connection.
        """
        await self._closed_event.wait()


@decorator.decorator
def _identity_decorator(func, *args, **kwargs):
    """Identity decorator.

    This isn't as useless as it sounds: given a function with a
    ``__signature__`` attribute, it generates a wrapper that really
    does have that signature.
    """
    return func(*args, **kwargs)


def wrap_handler(name: str, handler: Callable, fixed: int) -> Callable:
    """Convert a handler that takes a sequence of typed arguments into one
    that takes a message.

    The message is unpacked to the types given by the signature. If it could
    not be unpacked, the wrapper raises :exc:`FailReply`.

    Parameters
    ----------
    name
        Name of the message (only used to form error messages).
    handler
        The callable to wrap (may be a coroutine).
    fixed
        Number of leading parameters in `handler` that do not correspond to
        message arguments.
    """
    sig = inspect.signature(handler)
    pos = []
    var_pos = None
    for parameter in sig.parameters.values():
        if parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            var_pos = parameter
        elif parameter.kind in (inspect.Parameter.POSITIONAL_ONLY,
                                inspect.Parameter.POSITIONAL_OR_KEYWORD):
            pos.append(parameter)
        if parameter.name == '_msg':
            raise ValueError('Parameter cannot be named _msg')
    if len(pos) < fixed:
        raise TypeError('Handler must accept at least {} positional argument(s)'.format(fixed))

    # Exclude transferring __annotations__ from the wrapped function,
    # because the decorator does not preserve signature.
    @functools.wraps(handler, assigned=['__module__', '__name__', '__qualname__', '__doc__'])
    def wrapper(*args):
        assert len(args) == fixed + 1
        msg = args[-1]
        args = list(args[:-1])
        for argument in msg.arguments:
            if len(args) >= len(pos):
                if var_pos is None:
                    raise FailReply('too many arguments for {}'.format(name))
                else:
                    hint = var_pos.annotation
            else:
                hint = pos[len(args)].annotation
            if hint is inspect.Signature.empty:
                hint = bytes
            try:
                args.append(core.decode(hint, argument))
            except ValueError as error:
                raise FailReply(str(error)) from error
        try:
            return handler(*args)
        except TypeError as error:
            raise FailReply(str(error)) from error  # e.g. too few arguments

    if inspect.iscoroutinefunction(handler):
        wrapper = cast(Callable, types.coroutine(wrapper))

    wrapper_parameters = pos[:fixed]
    wrapper_parameters.append(
        inspect.Parameter('_msg', inspect.Parameter.POSITIONAL_OR_KEYWORD,
                          annotation=core.Message))
    wrapper.__signature__ = sig.replace(parameters=wrapper_parameters)  # type: ignore
    wrapper = _identity_decorator(wrapper)
    wrapper._aiokatcp_orig_handler = handler                            # type: ignore

    return wrapper



# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from .core import (                                              # noqa: F401
    Message, KatcpSyntaxError, Address, Timestamp, TimestampOrNow, Now,
    encode, decode, register_type, get_type, TypeInfo)
from .connection import Connection, FailReply, InvalidReply      # noqa: F401
from .server import DeviceServer, RequestContext                 # noqa: F401
from .client import (                                            # noqa: F401
    Client, ProtocolError, AbstractSensorWatcher, SensorWatcher,
    SyncState)
from .sensor import Reading, Sensor, SensorSampler, SensorSet    # noqa: F401


def minor_version():
    return '.'.join(__version__.split('.')[:2])

# Automatically added by katversion
__version__ = '0.8.0'



# Copyright 2017, 2019 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import asyncio
import re
import logging
import gc
import enum
import unittest
import unittest.mock
from unittest.mock import call
from typing import Tuple, Type, Pattern, Match, cast

from nose.tools import nottest, istest
import asynctest

from aiokatcp import (Client, FailReply, InvalidReply, ProtocolError, Message,
                      Sensor, SensorWatcher, AbstractSensorWatcher, SyncState, encode)
from .test_utils import timelimit


class DummyClient(Client):
    """Client with some informs for testing"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.foos = asyncio.Queue(loop=self.loop)
        self.unhandled = asyncio.Queue(loop=self.loop)

    def inform_foo(self, string: str, integer: int) -> None:
        self.foos.put_nowait((string, integer))

    def inform_exception(self) -> None:
        raise RuntimeError('I crashed')

    def unhandled_inform(self, msg: Message) -> None:
        self.unhandled.put_nowait(msg)


@nottest
class BaseTestClient(unittest.TestCase):
    async def make_server(self, loop: asyncio.AbstractEventLoop) \
            -> Tuple[asyncio.AbstractServer, asyncio.Queue]:
        """Start a server listening on localhost.

        Returns
        -------
        server
            Asyncio server
        client_queue
            Queue which is populated with `(reader, writer)` tuples as they connect
        """
        def callback(reader, writer):
            client_queue.put_nowait((reader, writer))

        client_queue = asyncio.Queue(loop=loop)   # type: asyncio.Queue
        server = await asyncio.start_server(callback, '127.0.0.1', 0, loop=loop)
        self.addCleanup(server.wait_closed)
        self.addCleanup(server.close)
        return server, client_queue


@timelimit
@nottest
class BaseTestClientAsync(BaseTestClient, asynctest.TestCase):
    async def make_client(
            self,
            server: asyncio.AbstractServer,
            client_queue: asyncio.Queue,
            client_cls: Type[Client] = DummyClient,
            auto_reconnect=True) \
            -> Tuple[Client, asyncio.StreamReader, asyncio.StreamWriter]:
        host, port = server.sockets[0].getsockname()    # type: ignore
        client = client_cls(host, port, auto_reconnect=auto_reconnect, loop=self.loop)
        self.addCleanup(client.wait_closed)
        self.addCleanup(client.close)
        (reader, writer) = await client_queue.get()
        return client, reader, writer

    async def check_received(self, data: bytes) -> None:
        line = await self.remote_reader.readline()
        self.assertEqual(line, data)

    async def check_received_regex(self, pattern: Pattern[bytes]) -> Match:
        line = await self.remote_reader.readline()
        self.assertRegex(line, pattern)
        # cast keeps mypy happy (it can't tell that it will always match after the assert)
        return cast(Match, pattern.match(line))

    async def write(self, data: bytes) -> None:
        self.remote_writer.write(data)
        await self.remote_writer.drain()

    async def wait_connected(self) -> None:
        self.remote_writer.write(b'#version-connect katcp-protocol 5.0-IM\n')
        await self.client.wait_connected()
        # Make sure that wait_connected works when already connected
        await self.client.wait_connected()


@timelimit
@istest
class TestClient(BaseTestClientAsync):
    @timelimit(1)
    async def setUp(self) -> None:
        self.server, self.client_queue = await self.make_server(self.loop)
        self.client, self.remote_reader, self.remote_writer = \
            await self.make_client(self.server, self.client_queue)
        self.addCleanup(self.remote_writer.close)

    async def test_request_ok(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('echo'))
        await self.check_received(b'?echo[1]\n')
        await self.write(b'!echo[1] ok\n')
        result = await future
        self.assertEqual(result, ([], []))
        # Again, with arguments. This also tests MID incrementing, non-ASCII
        # characters, and null escaping.
        arg = b'h\xaf\xce\0'
        arg_esc = b'h\xaf\xce\\0'  # katcp escaping
        future = self.loop.create_task(self.client.request('echo', b'123', arg))
        await self.check_received(b'?echo[2] 123 ' + arg_esc + b'\n')
        await self.write(b'!echo[2] ok 123 ' + arg_esc + b'\n')
        result = await future
        self.assertEqual(result, ([b'123', arg], []))

    async def test_request_fail(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('failme'))
        await self.check_received(b'?failme[1]\n')
        await self.write(b'!failme[1] fail Error\\_message\n')
        with self.assertRaisesRegex(FailReply, '^Error message$'):
            await future

    async def test_request_fail_no_msg(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('failme'))
        await self.check_received(b'?failme[1]\n')
        await self.write(b'!failme[1] fail\n')
        with self.assertRaisesRegex(FailReply, '^$'):
            await future

    async def test_request_fail_msg_bad_encoding(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('failme'))
        await self.check_received(b'?failme[1]\n')
        await self.write(b'!failme[1] fail \xaf\n')
        with self.assertRaisesRegex(FailReply, '^\uFFFD$'):
            await future

    async def test_request_invalid(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('invalid-request'))
        await self.check_received(b'?invalid-request[1]\n')
        await self.write(b'!invalid-request[1] invalid Unknown\\_request\n')
        with self.assertRaisesRegex(InvalidReply, '^Unknown request$'):
            await future

    async def test_request_no_code(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('invalid-request'))
        await self.check_received(b'?invalid-request[1]\n')
        await self.write(b'!invalid-request[1]\n')
        with self.assertRaisesRegex(InvalidReply, '^$'):
            await future

    async def test_request_with_informs(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('help'))
        await self.check_received(b'?help[1]\n')
        await self.write(b'#help[1] help Show\\_help\n')
        await self.write(b'#help[1] halt Halt\n')
        await self.write(b'!help[1] ok 2\n')
        result = await future
        self.assertEqual(result, ([b'2'], [
            Message.inform('help', b'help', b'Show help', mid=1),
            Message.inform('help', b'halt', b'Halt', mid=1)
        ]))

    async def test_inform(self) -> None:
        client = cast(DummyClient, self.client)
        await self.wait_connected()
        with self.assertLogs(logging.getLogger('aiokatcp.client')) as cm:
            # Put in bad ones before the good one, so that as soon as we've
            # received the good one from the queue we can finish the test.
            await self.write(b'#exception\n#foo bad notinteger\n#foo \xc3\xa9 123\n')
            inform = await client.foos.get()
        self.assertRegex(cm.output[0], 'I crashed')
        self.assertRegex(cm.output[1], 'error in inform')
        self.assertEqual(inform, ('é', 123))

    async def test_unhandled_inform(self) -> None:
        client = cast(DummyClient, self.client)
        await self.wait_connected()
        await self.write(b'#unhandled arg\n')
        msg = await client.unhandled.get()
        self.assertEqual(msg, Message.inform('unhandled', b'arg'))

    async def test_inform_callback(self) -> None:
        def callback(string: str, integer: int) -> None:
            values.put_nowait((string, integer))

        values = asyncio.Queue(loop=self.loop)    # type: asyncio.Queue[Tuple[str, int]]
        client = cast(DummyClient, self.client)
        client.add_inform_callback('bar', callback)
        await self.wait_connected()
        await self.write(b'#bar hello 42\n')
        value = await values.get()
        self.assertEqual(value, ('hello', 42))
        client.remove_inform_callback('bar', callback)
        self.assertEqual(client._inform_callbacks, {})

    async def test_unsolicited_reply(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('echo'))
        with self.assertLogs(logging.getLogger('aiokatcp.client'), logging.DEBUG):
            await self.write(b'!surprise[3]\n!echo[1] ok\n')
            await future

    async def test_receive_request(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('echo'))
        with self.assertLogs(logging.getLogger('aiokatcp.client')):
            await self.write(b'?surprise\n!echo[1] ok\n')
            await future

    async def test_reply_no_mid(self) -> None:
        await self.wait_connected()
        future = self.loop.create_task(self.client.request('echo'))
        with self.assertLogs(logging.getLogger('aiokatcp.client')):
            await self.write(b'!surprise ok\n!echo[1] ok\n')
            await future

    async def test_context_manager(self) -> None:
        async with self.client:
            pass
        await self.client.wait_closed()

    async def test_connect(self) -> None:
        host, port = self.server.sockets[0].getsockname()    # type: ignore
        client_task = self.loop.create_task(DummyClient.connect(host, port, loop=self.loop))
        (reader, writer) = await self.client_queue.get()
        await asynctest.exhaust_callbacks(self.loop)
        self.assertFalse(client_task.done())
        writer.write(b'#version-connect katcp-protocol 5.0-IM\n')
        client = await client_task
        assert client.is_connected
        client.close()
        writer.close()
        await client.wait_closed()

    async def test_unparsable_protocol(self) -> None:
        with self.assertLogs(logging.getLogger('aiokatcp.client')) as cm:
            self.remote_writer.write(b'#version-connect katcp-protocol notvalid\n')
            line = await self.remote_reader.read()
        self.assertEqual(line, b'')
        self.assertRegex(cm.output[0], 'Unparsable katcp-protocol')

    async def test_bad_protocol(self) -> None:
        with self.assertLogs(logging.getLogger('aiokatcp.client')) as cm:
            self.remote_writer.write(b'#version-connect katcp-protocol 4.0-I\n')
            line = await self.remote_reader.read()
        self.assertEqual(line, b'')
        self.assertRegex(cm.output[0], r'Unknown protocol version 4\.0')

    async def test_no_connection(self) -> None:
        # Open a second client, which will not get the #version-connect
        client, reader, writer = \
            await self.make_client(self.server, self.client_queue)
        self.addCleanup(writer.close)
        with self.assertRaises(BrokenPipeError):
            await client.request('help')

    async def test_connection_reset(self) -> None:
        await self.wait_connected()
        self.remote_writer.close()
        with self.assertRaises(ConnectionResetError):
            await self.client.request('help')

    async def test_disconnected(self) -> None:
        await self.wait_connected()
        await self.write(b'#disconnect Server\\_exiting\n')
        await self.client.wait_disconnected()
        with self.assertRaises(BrokenPipeError):
            await self.client.request('help')

    async def test_bad_address(self) -> None:
        client = DummyClient('invalid.invalid', 1)
        self.addCleanup(client.close)
        # While invalid.invalid will fail to resolve, it may take some time.
        # By mocking it, we ensure that the test runs without the need for
        # an estimated sleep.
        with unittest.mock.patch.object(self.loop, 'getaddrinfo', side_effect=OSError):
            with self.assertLogs(logging.getLogger('aiokatcp.client')) as cm:
                task = self.loop.create_task(client.wait_connected())
                await asynctest.exhaust_callbacks(self.loop)
        self.assertRegex(cm.output[0], 'Failed to connect to invalid.invalid:1: ')
        task.cancel()


@timelimit
@istest
class TestSensorMonitor(BaseTestClientAsync):
    """Test the sensor monitoring interface.

    This mocks out the :class:`~.AbstractSensorWatcher`.
    """

    @timelimit(1)
    async def setUp(self) -> None:
        self.server, self.client_queue = await self.make_server(self.loop)
        self.client, self.remote_reader, self.remote_writer = \
            await self.make_client(self.server, self.client_queue, auto_reconnect=True)
        self.addCleanup(self.remote_writer.close)
        self.watcher = unittest.mock.Mock(autospec=AbstractSensorWatcher)
        self.client.add_sensor_watcher(self.watcher)
        # Make it possible to wait for particular informs to reach the client
        # and get processed.
        self.wakeups = asyncio.Queue(loop=self.loop)  # type: asyncio.Queue[None]
        self.client.add_inform_callback('wakeup',
                                        lambda *args: self.wakeups.put_nowait(None))

    async def wakeup(self) -> None:
        """Wait for #wakeup sent from server to be received"""
        await self.wakeups.get()
        await asynctest.exhaust_callbacks(self.loop)

    async def connect(self) -> None:
        """Get as far as the monitor issuing ``?sensor-list``"""
        await self.wait_connected()
        self.watcher.state_updated.assert_called_with(SyncState.SYNCING)
        self.watcher.reset_mock()
        await self.check_received(b'?sensor-list[1]\n')

    async def sensor_list(self) -> None:
        """Send the sensor list and wait for ``?sensor-sampling``"""
        await self.write(
            b'#sensor-list[1] device-status Device\\_status \\@ discrete ok degraded fail\n'
            b'!sensor-list[1] ok 1\n')
        await self.check_received(b'?sensor-sampling[2] device-status auto\n')
        self.assertEqual(self.watcher.mock_calls, [
            call.batch_start(),
            call.sensor_added('device-status', 'Device status', '', 'discrete',
                              b'ok', b'degraded', b'fail'),
            call.batch_stop()
        ])
        self.watcher.reset_mock()

    async def sensor_sampling(self) -> None:
        """Reply to ``?sensor-sampling``"""
        await self.write(
            b'#sensor-status 123456789.0 1 device-status nominal ok\n'
            b'!sensor-sampling[2] ok device-status auto\n'
            b'#wakeup\n')
        await self.wakeup()
        self.assertEqual(self.watcher.mock_calls, [
            call.batch_start(),
            call.sensor_updated('device-status', b'ok', Sensor.Status.NOMINAL,
                                123456789.0),
            call.batch_stop(),
            call.state_updated(SyncState.SYNCED)
        ])
        self.watcher.reset_mock()

    async def test_init(self) -> None:
        await self.connect()
        await self.sensor_list()
        await self.sensor_sampling()

    async def interface_changed(self):
        """Send a ``#interface-changed`` inform and wait for ``?sensor-list``"""
        await self.write(b'#interface-changed sensor-list\n#wakeup\n')
        await self.wakeup()
        self.watcher.state_updated.assert_called_with(SyncState.SYNCING)
        self.watcher.reset_mock()
        await self.check_received(b'?sensor-list[3]\n')

    async def test_add_remove_sensors(self):
        await self.test_init()
        await self.interface_changed()
        await self.write(
            b'#sensor-list[3] temp Temperature F float\n'
            b'!sensor-list[3] ok 1\n')
        await self.check_received(b'?sensor-sampling[4] temp auto\n')
        self.assertEqual(self.watcher.mock_calls, [
            call.batch_start(),
            call.sensor_added('temp', 'Temperature', 'F', 'float'),
            call.sensor_removed('device-status'),
            call.batch_stop()
        ])
        self.watcher.reset_mock()

        await self.write(
            b'#sensor-status 123456790.0 1 temp warn 451.0\n'
            b'!sensor-sampling[4] ok temp auto\n'
            b'#wakeup\n')
        await self.wakeup()
        self.assertEqual(self.watcher.mock_calls, [
            call.batch_start(),
            call.sensor_updated('temp', b'451.0', Sensor.Status.WARN, 123456790.0),
            call.batch_stop(),
            call.state_updated(SyncState.SYNCED)
        ])
        self.watcher.reset_mock()

    async def test_replace_sensor(self):
        """Sensor has the same name but different parameters"""
        await self.test_init()
        await self.interface_changed()
        await self.write(
            b'#sensor-list[3] device-status A\\_different\\_status \\@ int\n'
            b'!sensor-list[3] ok 1\n')
        await self.check_received(b'?sensor-sampling[4] device-status auto\n')
        self.assertEqual(self.watcher.mock_calls, [
            call.batch_start(),
            call.sensor_added('device-status', 'A different status', '', 'int'),
            call.batch_stop()
        ])
        self.watcher.reset_mock()

        await self.write(
            b'#sensor-status 123456791.0 1 device-status nominal 123\n'
            b'!sensor-sampling[4] ok device-status auto\n'
            b'#wakeup\n')
        await self.wakeup()
        self.assertEqual(self.watcher.mock_calls, [
            call.batch_start(),
            call.sensor_updated('device-status', b'123', Sensor.Status.NOMINAL, 123456791.0),
            call.batch_stop(),
            call.state_updated(SyncState.SYNCED)
        ])
        self.watcher.reset_mock()

    async def test_sensor_vanished(self):
        """Sensor vanishes immediately after sensor-list reply."""
        await self.connect()
        await self.sensor_list()
        await self.write(
            b'#interface-changed sensor-list\n'
            b"!sensor-sampling[2] fail Unknown\\_sensor\\_'device-status'\n")
        await self.check_received(b'?sensor-list[3]\n')
        await self.write(b'!sensor-list[3] ok 0\n#wakeup\n')
        await self.wakeup()
        self.assertEqual(self.watcher.mock_calls, [
            call.state_updated(SyncState.SYNCING),
            call.batch_start(),
            call.sensor_removed('device-status'),
            call.batch_stop(),
            call.state_updated(SyncState.SYNCED)
        ])
        self.watcher.reset_mock()

    async def test_sensor_vanished2(self):
        """Sensor vanishes immediately after sensor-list reply (second case).

        This is similar to :meth:`test_sensor_vanished`, but the inform arrives
        only after the failure in ``?sensor-sampling``.
        """
        await self.connect()
        await self.sensor_list()
        await self.write(
            b"!sensor-sampling[2] fail Unknown\\_sensor\\_'device-status'\n"
            b'#wakeup\n')
        # Wait until the update task finishes before sending interface-changed
        await self.wakeup()
        await self.write(b'#interface-changed sensor-list\n')
        await self.check_received(b'?sensor-list[3]\n')
        await self.write(b'!sensor-list[3] ok 0\n#wakeup\n')
        await self.wakeup()
        self.assertEqual(self.watcher.mock_calls, [
            call.state_updated(SyncState.SYNCED),
            call.state_updated(SyncState.SYNCING),
            call.batch_start(),
            call.sensor_removed('device-status'),
            call.batch_stop(),
            call.state_updated(SyncState.SYNCED)
        ])
        self.watcher.reset_mock()

    async def test_remove_sensor_watcher(self):
        """Removing the last watcher unsubscribes"""
        await self.test_init()
        self.client.remove_sensor_watcher(self.watcher)
        await self.check_received(b'?sensor-sampling[3] device-status none\n')
        await self.write(b'!sensor-sampling[3] ok device-status none\n')

    async def test_close(self):
        """Closing the client must update the state"""
        await self.test_init()
        self.client.close()
        await self.client.wait_closed()
        self.assertEqual(self.watcher.mock_calls, [
            call.state_updated(SyncState.CLOSED)
        ])

    async def test_disconnect(self):
        """When the connection drops, the state must change appropriately"""
        await self.test_init()
        await self.write(b'#disconnect Testing\n')
        (self.remote_reader, self.remote_writer) = await self.client_queue.get()

        await self.wait_connected()
        self.watcher.state_updated.assert_called_with(SyncState.SYNCING)
        self.watcher.reset_mock()
        await self.check_received(b'?sensor-list[3]\n')
        await self.write(
            b'#sensor-list[3] device-status Device\\_status \\@ discrete ok degraded fail\n'
            b'!sensor-list[3] ok 1\n')
        await self.check_received(b'?sensor-sampling[4] device-status auto\n')
        await self.write(
            b'#sensor-status 123456789.0 1 device-status nominal ok\n'
            b'!sensor-sampling[4] ok device-status auto\n'
            b'#wakeup\n')
        await self.wakeup()
        self.assertEqual(self.watcher.mock_calls, [
            call.batch_start(),
            # No sensor_added because the sensor was already known
            call.batch_stop(),
            call.batch_start(),
            call.sensor_updated('device-status', b'ok', Sensor.Status.NOMINAL, 123456789.0),
            call.batch_stop(),
            call.state_updated(SyncState.SYNCED)
        ])
        self.watcher.reset_mock()


class DummySensorWatcher(SensorWatcher):
    def rewrite_name(self, name: str) -> str:
        return 'test_' + name


class DummyEnum(enum.Enum):
    THING_ONE = 1
    THING_TWO = 2


class TestSensorWatcher(asynctest.TestCase):
    """Test :class:`~.SensorWatcher`."""
    def setUp(self):
        client = unittest.mock.MagicMock()
        client.loop = self.loop
        self.watcher = DummySensorWatcher(client, enum_types=[DummyEnum])

    def test_construct(self):
        self.assertEqual(len(self.watcher.sensors), 0)
        self.assertFalse(self.watcher.synced.is_set())

    def test_sensor_added(self):
        self.watcher.batch_start()
        self.watcher.sensor_added('foo', 'A sensor', 'F', 'float')
        self.watcher.batch_stop()
        self.assertEqual(len(self.watcher.sensors), 1)
        sensor = self.watcher.sensors['test_foo']
        self.assertEqual(sensor.name, 'test_foo')
        self.assertEqual(sensor.description, 'A sensor')
        self.assertEqual(sensor.units, 'F')
        self.assertEqual(sensor.stype, float)
        self.assertEqual(sensor.status, Sensor.Status.UNKNOWN)

    def test_sensor_added_discrete(self):
        self.watcher.batch_start()
        self.watcher.sensor_added('disc', 'Discrete sensor', '', 'discrete', b'abc', b'def-xyz')
        self.watcher.sensor_added('disc2', 'Discrete sensor 2', '', 'discrete', b'abc', b'def-xyz')
        self.watcher.batch_stop()
        self.assertEqual(len(self.watcher.sensors), 2)
        sensor = self.watcher.sensors['test_disc']
        self.assertEqual(sensor.name, 'test_disc')
        self.assertEqual(sensor.description, 'Discrete sensor')
        self.assertEqual(sensor.units, '')
        self.assertEqual(sensor.type_name, 'discrete')
        self.assertEqual(sensor.status, Sensor.Status.UNKNOWN)
        members = [encode(member) for member in sensor.stype.__members__.values()]
        self.assertEqual(members, [b'abc', b'def-xyz'])
        self.assertIs(self.watcher.sensors['test_disc'].stype,
                      self.watcher.sensors['test_disc2'].stype,
                      'Enum cache did not work')

    def test_sensor_added_known_discrete(self):
        self.watcher.batch_start()
        self.watcher.sensor_added('disc', 'Discrete sensor', '', 'discrete',
                                  b'thing-one', b'thing-two')
        self.watcher.batch_stop()
        self.assertEqual(len(self.watcher.sensors), 1)
        sensor = self.watcher.sensors['test_disc']
        self.assertEqual(sensor.name, 'test_disc')
        self.assertEqual(sensor.description, 'Discrete sensor')
        self.assertEqual(sensor.units, '')
        self.assertEqual(sensor.type_name, 'discrete')
        self.assertIs(sensor.stype, DummyEnum)
        self.assertEqual(sensor.status, Sensor.Status.UNKNOWN)

    def test_sensor_added_bad_type(self):
        self.watcher.batch_start()
        self.watcher.sensor_added('foo', 'A sensor', 'F', 'blah')
        self.watcher.batch_stop()
        self.assertEqual(len(self.watcher.sensors), 0)
        self.watcher.logger.warning.assert_called_once_with(
            'Type %s is not recognised, skipping sensor %s', 'blah', 'foo')

    def test_sensor_removed(self):
        self.test_sensor_added()

        self.watcher.batch_start()
        self.watcher.sensor_removed('foo')
        self.watcher.batch_stop()
        self.assertEqual(len(self.watcher.sensors), 0)

    def test_sensor_updated(self):
        self.test_sensor_added()

        self.watcher.batch_start()
        self.watcher.sensor_updated('foo', b'12.5', Sensor.Status.WARN, 1234567890.0)
        self.watcher.batch_stop()
        sensor = self.watcher.sensors['test_foo']
        self.assertEqual(sensor.value, 12.5)
        self.assertEqual(sensor.status, Sensor.Status.WARN)
        self.assertEqual(sensor.timestamp, 1234567890.0)

    def test_sensor_updated_bad_value(self):
        self.test_sensor_added()

        self.watcher.batch_start()
        self.watcher.sensor_updated('foo', b'not a float', Sensor.Status.WARN, 1234567890.0)
        self.watcher.batch_stop()
        self.watcher.logger.warning.assert_called_once_with(
            'Sensor %s: value %r does not match type %s: %s',
            'foo', b'not a float', 'float', unittest.mock.ANY)

    def test_sensor_updated_unknown_sensor(self):
        self.test_sensor_added()

        self.watcher.batch_start()
        self.watcher.sensor_updated('bar', b'123.0', Sensor.Status.WARN, 1234567890.0)
        self.watcher.batch_stop()
        self.watcher.logger.warning.assert_called_once_with(
            'Received update for unknown sensor %s', 'bar')

    def test_state_updated(self):
        self.test_sensor_added()

        self.watcher.state_updated(SyncState.SYNCING)
        self.assertFalse(self.watcher.synced.is_set())
        self.assertEqual(len(self.watcher.sensors), 1)
        self.assertEqual(self.watcher.sensors['test_foo'].status, Sensor.Status.UNKNOWN)

        self.watcher.state_updated(SyncState.SYNCED)
        self.assertTrue(self.watcher.synced.is_set())
        self.assertEqual(len(self.watcher.sensors), 1)
        self.assertEqual(self.watcher.sensors['test_foo'].status, Sensor.Status.UNKNOWN)

        self.watcher.state_updated(SyncState.DISCONNECTED)
        self.assertFalse(self.watcher.synced.is_set())
        # Disconnecting should set all sensors to UNREACHABLE
        self.assertEqual(self.watcher.sensors['test_foo'].status, Sensor.Status.UNREACHABLE)


@timelimit
@istest
class TestClientNoReconnect(TestClient):
    @timelimit(1)
    async def setUp(self) -> None:
        self.server, self.client_queue = await self.make_server(self.loop)
        self.client, self.remote_reader, self.remote_writer = \
            await self.make_client(self.server, self.client_queue, auto_reconnect=False)
        self.addCleanup(self.remote_writer.close)

    async def test_unparsable_protocol(self) -> None:
        self.remote_writer.write(b'#version-connect katcp-protocol notvalid\n')
        line = await self.remote_reader.read()
        self.assertEqual(line, b'')
        with self.assertRaises(ProtocolError):
            await self.client.wait_connected()

    async def test_bad_protocol(self) -> None:
        # Different approach to test_unparsable_protocol, to get more coverage
        wait_task = self.loop.create_task(self.client.wait_connected())
        self.remote_writer.write(b'#version-connect katcp-protocol 4.0-I\n')
        line = await self.remote_reader.read()
        self.assertEqual(line, b'')
        with self.assertRaises(ProtocolError):
            await wait_task

    async def test_disconnected(self) -> None:
        await self.wait_connected()
        await self.write(b'#disconnect Server\\_exiting\n')
        await self.client.wait_disconnected()
        with self.assertRaises(BrokenPipeError):
            await self.client.request('help')
        with self.assertRaises(ConnectionResetError):
            await self.client.wait_connected()

    async def test_connect_failed(self) -> None:
        host, port = self.server.sockets[0].getsockname()    # type: ignore
        client_task = self.loop.create_task(
            DummyClient.connect(host, port, auto_reconnect=False, loop=self.loop))
        (reader, writer) = await self.client_queue.get()
        await asynctest.exhaust_callbacks(self.loop)
        self.assertFalse(client_task.done())
        writer.close()
        with self.assertRaises(ConnectionAbortedError):
            await client_task


@timelimit
@istest
class TestClientNoMidSupport(BaseTestClientAsync):
    @timelimit(1)
    async def setUp(self) -> None:
        self.server, self.client_queue = await self.make_server(self.loop)
        self.client, self.remote_reader, self.remote_writer = \
            await self.make_client(self.server, self.client_queue)
        self.addCleanup(self.remote_writer.close)

    async def test_single(self):
        self.remote_writer.write(b'#version-connect katcp-protocol 5.0-M\n')
        await self.client.wait_connected()
        future = self.loop.create_task(self.client.request('echo'))
        await self.check_received(b'?echo\n')
        await self.write(b'#echo an\\_inform\n')
        await self.write(b'!echo ok\n')
        result = await future
        self.assertEqual(result, ([], [Message.inform('echo', b'an inform')]))

    async def test_concurrent(self):
        self.remote_writer.write(b'#version-connect katcp-protocol 5.0-M\n')
        await self.client.wait_connected()
        future1 = self.loop.create_task(self.client.request('echo', 1))
        future2 = self.loop.create_task(self.client.request('echo', 2))
        for i in range(2):
            match = await self.check_received_regex(re.compile(br'^\?echo (1|2)\n\Z'))
            await self.write(b'#echo value ' + match.group(1) + b'\n')
            await self.write(b'!echo ok ' + match.group(1) + b'\n')
        result1 = await future1
        self.assertEqual(result1, ([b'1'], [Message.inform('echo', b'value', b'1')]))
        result2 = await future2
        self.assertEqual(result2, ([b'2'], [Message.inform('echo', b'value', b'2')]))


@istest
class TestUnclosedClient(BaseTestClient, unittest.TestCase):
    async def body(self) -> None:
        server, client_queue = await self.make_server(self.loop)
        host, port = server.sockets[0].getsockname()    # type: ignore
        client = DummyClient(host, port, loop=self.loop)  # noqa: F841
        (reader, writer) = await client_queue.get()
        writer.close()
        server.close()
        await server.wait_closed()

    def test(self) -> None:
        with self.assertWarnsRegex(ResourceWarning, 'unclosed Client'):
            self.loop = asyncio.new_event_loop()
            self.loop.run_until_complete(self.body())
            self.loop.close()
            # Run a few times for PyPy's benefit
            gc.collect()
            gc.collect()



# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools
import inspect
import sys
import asyncio

import async_timeout


def timelimit(limit=5.0):
    """Decorator to run tests with a time limit. It is designed to be used
    with :class:`asynctest.TestCase`. It can be used as either a method or
    a class decorator. It can be used as either ``@timelimit(limit)`` or
    just ``@timelimit`` to use the default of 5 seconds.
    """
    if inspect.isfunction(limit) or inspect.isclass(limit):
        # Used without parameters
        return timelimit()(limit)

    def decorator(arg):
        if inspect.isclass(arg):
            for key, value in arg.__dict__.items():
                if (inspect.iscoroutinefunction(value) and key.startswith('test_')
                        and not hasattr(arg, '_timelimit')):
                    setattr(arg, key, decorator(value))
            return arg
        else:
            @functools.wraps(arg)
            async def wrapper(self, *args, **kwargs):
                try:
                    async with async_timeout.timeout(limit, loop=self.loop) as cm:
                        await arg(self, *args, **kwargs)
                except asyncio.TimeoutError:
                    if not cm.expired:
                        raise
                    for task in asyncio.Task.all_tasks(loop=self.loop):
                        if task.get_stack(limit=1):
                            print()
                            task.print_stack(file=sys.stdout)
                    raise asyncio.TimeoutError('Test did not complete within {}s'.format(limit))
            wrapper._timelimit = limit
            return wrapper
    return decorator
